Package: packFinder
Type: Package
Title: de novo Annotation of Pack-TYPE Transposable Elements
Version: 1.25.0
Authors@R: c(
    person(given = "Jack", 
           family = "Gisby", 
           role = c("aut", "cre"), 
           email = "jackgisby@gmail.com",
           comment = c(ORCID = "0000-0003-0511-8123")), 
    person(given = "Marco", 
           family = "Catoni", 
           role = "aut", 
           email="m.catoni@bham.ac.uk",
           comment = c(ORCID = "0000-0002-3258-2522"))
    )
Description: Algorithm and tools for in silico pack-TYPE transposon discovery. Filters a given genome for properties unique to DNA transposons and provides tools for the investigation of returned matches. Sequences are input in DNAString format, and ranges are returned as a dataframe (in the format returned by as.dataframe(GRanges)).
License: GPL-2
Encoding: UTF-8
Depends: R (>= 4.1.0)
Imports: Biostrings, GenomicRanges, kmer, ape, methods, IRanges,
        S4Vectors
Suggests: biomartr, knitr, rmarkdown, testthat, dendextend, biocViews,
        BiocCheck, BiocStyle
RoxygenNote: 7.2.0
VignetteBuilder: knitr
URL: https://github.com/jackgisby/packFinder
BugReports: https://github.com/jackgisby/packFinder/issues
biocViews: Genetics, SequenceMatching, Annotation
git_url: https://git.bioconductor.org/packages/packFinder
git_branch: devel
git_last_commit: bf96429
git_last_commit_date: 2026-04-28
Repository: Bioconductor 3.24
Date/Publication: 2026-04-28
NeedsCompilation: no
Packaged: 2026-04-29 00:12:21 UTC; biocbuild
Author: Jack Gisby [aut, cre] (ORCID: <https://orcid.org/0000-0003-0511-8123>),
  Marco Catoni [aut] (ORCID: <https://orcid.org/0000-0002-3258-2522>)
Maintainer: Jack Gisby <jackgisby@gmail.com>

# Generated by roxygen2: do not edit by hand

export(blastAnalysis)
export(blastAnnotate)
export(collapseSeqs)
export(filterWildcards)
export(getPackSeqs)
export(getPacksFromCsv)
export(getPacksFromFasta)
export(getPacksFromGRanges)
export(getTsds)
export(identifyPotentialPackElements)
export(identifyTirMatches)
export(makeBlastDb)
export(packAlign)
export(packBlast)
export(packClust)
export(packSearch)
export(packsToCsv)
export(packsToFasta)
export(packsToGRanges)
export(readBlast)
export(readUc)
export(tirClust)

# packFinder <img src="inst/packFinder_hex.png" align="right" height="174" width="150" />

A package for the *de novo* discovery of Pack-TYPE transposons. Transposons are detected in DNA sequences based on conserved terminal inverted repeat sequences and presence of terminal site duplications. `packFinder` allows users to search a given Genome for Pack-TYPE transposons with minimal input and setup. For a more in-depth explanation, see our [paper in PLOS Genetics](https://doi.org/10.1371/journal.pgen.1010078).

## Installation
`packFinder` is available from the [Bioconductor project](https://bioconductor.org/packages/packFinder/). With R >=4.0.0 installed, the following code can be used to install the package:
```
if (!require("BiocManager"))
    install.packages("BiocManager")
    
BiocManager::install("packFinder")
```

In order to use the functions `packClust` and `packAlign`, the VSEARCH command line tool must be installed. For Linux and MacOS systems, correct installation of VSEARCH should allow users to use all functions within `packFinder`; for windows users, the absolute path to the VSEARCH executable file must be specified when calling `packFinder` clustering and alignment functions. See the vignette, documentation or the [VSEARCH github](https://github.com/torognes/vsearch) for further information.

Additionally, utilites are provided with the package for the functional annotation of internal transposon sequences using BLAST+; this application can be installed from [NCBI](https://blast.ncbi.nlm.nih.gov/Blast.cgi?PAGE_TYPE=BlastDocs&DOC_TYPE=Download).

## Using packFinder
The `packFinder` vignette includes a full walkthrough of the package. To get started quickly, it is easiest to download a genome of interest using the `biomartr` package and follow the steps outlined in [the vignette](https://bioconductor.org/packages/release/bioc/vignettes/packFinder/inst/doc/packFinder.html).

## Citation
If you use `packFinder` in your work, please cite the following publication:

> Gisby, J. S. & Catoni, M. (2022) [The widespread nature of Pack-TYPE transposons reveals their importance for plant genome evolution.](https://doi.org/10.1371/journal.pgen.1010078) *PLOS Genetics* 18(2): e1010078.
> 

## ----setup, include = FALSE---------------------------------------------------
BiocStyle::markdown()

library(packFinder)

dir.create("tempOutput")

## ----packFinderInstall, eval=FALSE--------------------------------------------
# if (!requireNamespace("BiocManager", quietly = TRUE))
#     install.packages("BiocManager")
# 
# # The following initializes usage of Bioc devel
# BiocManager::install(version='devel')
# 
# BiocManager::install("packFinder")

## ----biostrings---------------------------------------------------------------
# Convert a character vector to a DNAString
Biostrings::DNAString("CATG")

# Convert a list of character vectors to a DNAStringSet
Biostrings::DNAStringSet(c(
    "CATG",
    "GTAC"
))

# Convert a FASTA file to a DNAStringSet
Biostrings::readDNAStringSet(
    system.file("extdata", "packMatches.fasta", package = "packFinder"),
    format = "fasta"
)

## ----packSearch---------------------------------------------------------------
data("arabidopsisThalianaRefseq")

packMatches <- packSearch(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    elementLength = c(300, 3500),
    tsdLength = 3
)

head(packMatches)

## ----packClustData, include=FALSE---------------------------------------------
# We don't have the dependencies for clustering, so get pre-clustered results
data(packMatches)

## ----packClust, eval=FALSE----------------------------------------------------
# packMatches <- packClust(
#     packMatches,
#     arabidopsisThalianaRefseq,
#     saveFolder = "tempOutput"
# )

## ----readClust----------------------------------------------------------------
readUc(system.file(
    "extdata",
    "packMatches.uc",
    package = "packFinder"
))

readBlast(system.file(
    "extdata",
    "packMatches.blast6out",
    package = "packFinder"
))

## ----tirClust-----------------------------------------------------------------
consensusSeqs <- tirClust(packMatches,
    arabidopsisThalianaRefseq,
    tirLength = 25
)

head(consensusSeqs)

## ----align, eval=FALSE--------------------------------------------------------
# packMatches <- packAlign(
#     packMatches,
#     arabidopsisThalianaRefseq,
#     saveFolder = "tempOutput"
# )

## ----blast--------------------------------------------------------------------
# the packMatches dataframe is exported as a FASTA file for NCBI blast search
packsToFasta(
    packMatches,
    "tempOutput/packMatches.fasta",
    arabidopsisThalianaRefseq
)

## ----csvConvert---------------------------------------------------------------
packsToCsv(packMatches, "tempOutput/packMatches.csv")
print(getPacksFromCsv("tempOutput/packMatches.csv"))

## ----grangesConvert-----------------------------------------------------------
packsGRanges <- packsToGRanges(packMatches)
print(packsGRanges)

print(getPacksFromGRanges(packsGRanges))

## ----fastaConvert-------------------------------------------------------------
packsToFasta(
    packMatches,
    "tempOutput/packMatches.fasta",
    arabidopsisThalianaRefseq
)

print(getPacksFromFasta("tempOutput/packMatches.fasta"))

## ----identifyTirMatches-------------------------------------------------------
forwardMatches <- identifyTirMatches(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    strand = "+",
    tsdLength = 3
)
nrow(forwardMatches)

reverseMatches <- identifyTirMatches(
    Biostrings::reverseComplement(Biostrings::DNAString("CACTACAA")),
    arabidopsisThalianaRefseq,
    strand = "-",
    tsdLength = 3
)
nrow(reverseMatches)

## ----getTsds------------------------------------------------------------------
forwardMatches$TSD <- getTsds(
    forwardMatches,
    arabidopsisThalianaRefseq,
    3,
    strand = "+"
)

head(forwardMatches)

reverseMatches$TSD <- getTsds(
    reverseMatches,
    arabidopsisThalianaRefseq,
    3,
    strand = "-"
)

head(reverseMatches)

## ----identifyPotentialPackElements--------------------------------------------
identifyPotentialPackElements(
    forwardMatches,
    reverseMatches,
    arabidopsisThalianaRefseq,
    c(300, 3500)
)

## ----getPackSeqs--------------------------------------------------------------
getPackSeqs(packMatches, arabidopsisThalianaRefseq)

## ----SessionInfo--------------------------------------------------------------
sessionInfo()

## ----include=FALSE------------------------------------------------------------
unlink("tempOutput", recursive = TRUE)

# this script was used to generate the file found at:
# inst/extdata/packMatches.csv

library(packFinder)

data("packMatches")

packsToCsv(packMatches, system.file(
    "extdata", 
    "packMatches.csv", 
    package = "packFinder"
))

# this script was used to generate the file found at:
# inst/extdata/packMatches.fasta

library(packFinder)

data("arabidopsisThalianaRefseq")
data("packMatches")

packsToFasta(packMatches, system.file(
    "extdata", 
    "packMatches.fasta", 
    package = "packFinder"
), arabidopsisThalianaRefseq)

# this script was used to generate the file found at:
# inst/extdata/packMatches.blast6out

library(packFinder)

data("arabidopsisThalianaRefseq")
data("packMatches")

packClusts <- packClust(
    packMatches, 
    arabidopsisThalianaRefseq,
    saveFolder = system.file("extdata", package = "packFinder")
)

#' @title
#' Pipeline for BLAST/Classification of PackTYPE Elements
#'
#' @description
#' Run BLAST against user-specified databases of 
#' non-transposon and transposon-relates proteins.
#' Can be used to classify transposons based on 
#' their internal sequences. 
#'
#' @param packMatches
#' A dataframe of potential Pack-TYPE transposable elements, 
#' in the format given by \code{\link{packSearch}}. This 
#' dataframe is in the format produced by coercing a 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' Will be saved as a FASTA file for VSEARCH.
#'
#' @param Genome
#' A DNAStringSet object containing sequences referred to 
#' in \code{packMatches} (the object originally used to 
#' predict the transposons \code{\link{packSearch}}). 
#' 
#' @param blastPath
#' Path to the BLAST+ executable, or name of 
#' the BLAST+ application for Linux/MacOS users.
#' 
#' @param protDb
#' For assigning Pack-TYPE elements. 
#' Path to the blast database containing nucleotide or protein
#' sequences to be matched against internal transposon 
#' sequences. Can be generated 
#' using BLAST+, or with 
#' \code{link{makeBlastDb}}.
#' 
#' @param autoDb
#' For assigning autonomous elements. 
#' Path to the blast database containing nucleotide or protein
#' sequences to be matched against internal transposon 
#' sequences. Can be generated 
#' using BLAST+, or with 
#' \code{link{makeBlastDb}}.
#' 
#' @param minE
#' Blast results with e values greater than
#' the specified cutoff will be ignored. This will 
#' be passed to BLASTN and applied to both transposon 
#' and non-transposon matches.
#' 
#' @param blastTask
#' Type of BLAST+ task, defaults to "blastn-short".
#' 
#' @param maxHits
#' Maximum hits returned by BLAST+ per query.
#' 
#' @param threads
#' Allowable number of threads to be utilised by BLAST+.
#' 
#' @param saveFolder
#' Directory to save BLAST+ results in; defaults 
#' to the working directory.
#' 
#' @param tirCutoff
#' How many bases to ignore at the terminal ends of the 
#' transposons to prevent hits to TIR sequences.
#' 
#' @param autoCutoff
#' Blast results for transposon-related elements will be 
#' filtered to 
#' ignore those with e values above the specified cutoff.
#' 
#' @param autoLength
#' Blast results for transposon-related elements containing 
#' hits with 
#' alignment lengths lower than this value will be ignored
#' 
#' @param autoIdentity
#' Blast results for transposon-related elements containing 
#' hits with 
#' sequence identities lower than this value will be ignored
#' 
#' @param autoScope
#' If specified, transposon-related blast results below the 
#' specified value
#' will be ignored. Note that the dataframe of transposon
#' matches must also be supplied to calculate scope. Scope is 
#' the proportion of the transposon's internal sequence 
#' occupied by the BLAST hit. 
#' 
#' @param protCutoff
#' Blast results for genic/other matches will be 
#' filtered to 
#' ignore those with e values above the specified cutoff.
#' 
#' @param protLength
#' Blast results for genic/other matches containing 
#' hits with 
#' alignment lengths lower than this value will be ignored
#' 
#' @param protIdentity
#' Blast results for genic/other matches containing 
#' hits with 
#' sequence identities lower than this value will be ignored
#' 
#' @param protScope
#' If specified, genic/other blast matches below the 
#' specified value
#' will be ignored. Note that the dataframe of transposon
#' matches must also be supplied to calculate scope. Scope is 
#' the proportion of the transposon's internal sequence 
#' occupied by the BLAST hit. 
#' 
#' @return 
#' Returns the original \code{packMatches} dataframe, 
#' with the addition of a "classification" column 
#' containing one of the following values:
#' \itemize{
#'     \item auto - elements that match known 
#'     transposases or transposon-related proteins 
#'     are classified as autonomous elements
#'     \item pack - elements that match other 
#'     proteins or genic sequences may be classified 
#'     as Pack-TYPE elements
#'     \item other - elements that generate no 
#'     significant hits
#' }
#'     
#' @seealso 
#' \code{\link{blastAnalysis}}, \code{\link{packSearch}},
#' \code{\link{readBlast}}, \code{\link{blastAnnotate}}
#' 
#' @examples
#' \dontrun{
#' packMatches <- data(packMatches)
#' Genome <- data(arabidopsisThalianaRefseq)
#' 
#' packBlast(packMatches, Genome, 
#'     protDb = "C:/data/TAIR10_CDS", 
#'     autoDb = "C:/data/TAIR10_transposons", 
#'     blastPath = "C:/blast/bin/blastn.exe")
#' }
#' 
#' @references 
#' For further information, see the NCBI BLAST+ application
#' documentation and help pages 
#' (https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
#'  
#' @author
#' Jack Gisby
#'
#' @export

packBlast <- function(packMatches, Genome, blastPath, protDb, autoDb,
                      minE = 1e-3, blastTask = "blastn-short", maxHits = 100,
                      threads = 1, saveFolder = NULL, tirCutoff = 100,
                      autoCutoff = 1e-5, autoLength = 150, autoIdentity = 70,
                      autoScope = NULL, protCutoff = 1e-5, protLength = 250, 
                      protIdentity = 70, protScope = 0.3) {
    
    if (is.null(saveFolder)) {
        saveFolder <- getwd()
    }
    
    blastAnalysis(packMatches, Genome, protDb = protDb, autoDb = autoDb, 
                  blastPath = blastPath, minE = minE, blastTask = blastTask,
                  maxHits = maxHits, threads = threads, saveFolder = saveFolder,
                  tirCutoff = tirCutoff)
    
    packMatches$id <- rownames(packMatches)
    
    autoHits <- readBlast(file.path(saveFolder, "autoHits.blast"), 
                          minE = autoCutoff, length = autoLength, 
                          identity = autoIdentity, removeExactMatches = FALSE, 
                          scope = autoScope, packMatches = packMatches)
    
    protHits <- readBlast(file.path(saveFolder, "protHits.blast"), 
                          minE = protCutoff, length = protLength, 
                          identity = protIdentity, removeExactMatches = TRUE,
                          scope = protScope, packMatches = packMatches)
    
    return(blastAnnotate(protHits, autoHits, packMatches))
}

#' @title
#' Identify Terminal Inverted Repeat Matches
#'
#' @description
#' Searches a \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#' for potential TIRs based on sequence similarity.
#'
#' @param tirSeq 
#' A \code{\link[Biostrings:DNAString-class]{DNAString}} 
#' object to be searched for.
#'
#' @param Genome
#' A \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#' object containing the 
#' \code{\link[Biostrings:DNAString-class]{DNAString}} 
#' objects to be searched.
#'
#' @param mismatch
#' The allowable mismatch between \code{tirSeq} and a 
#' given slice of \code{Genome}. Includes indels.
#'
#' @param strand
#' The directionality of the search string ("+" or "-"). Note
#' that this does affect the search for tirSeqs, if you wish 
#' to search the reverse strand you should use the reverse 
#' complement of your sequence. 
#'
#' @param tsdLength
#' Integer referring to the length of the flanking TSD region.
#' 
#' @param fixed
#' Logical that will be passed to the `fixed` argument of 
#' \code{\link[Biostrings]{matchPattern}}. Determines the behaviour of IUPAC
#' ambiguity codes when searching for TIR sequences.
#'
#' @details 
#' Called by \code{\link{packSearch}}. Used by 
#' \code{\link{packSearch}} as an initial filtering stage. 
#' \code{\link[Biostrings]{matchPattern}} from Biostrings 
#' is used for pattern matching. It is recommended to use 
#' the general pipeline function \code{\link{packSearch}} 
#' for identification of potential pack elements, however 
#' each stage may be called individually.
#'
#' @return
#' A dataframe, \code{tirMatches}, containing identified 
#' matches. The dataframe is in the format generated by 
#' \code{\link{packSearch}}.
#' 
#' @seealso 
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
#' \code{\link{packSearch}}, \code{\link[Biostrings]{matchPattern}},
#' \code{\link[Biostrings:DNAString-class]{DNAString}} 
#' 
#' @examples
#' data(arabidopsisThalianaRefseq)
#' 
#' forwardMatches <- identifyTirMatches(
#'     Biostrings::DNAString("CACTACAA"),
#'     arabidopsisThalianaRefseq,
#'     tsdLength = 3,
#'     strand = "+"
#' )
#' 
#' @author
#' Jack Gisby
#'
#' @export

identifyTirMatches <- function(tirSeq, Genome, mismatch = 0, strand = "*", 
                               tsdLength, fixed=TRUE) {
    if (strand != "-" & strand != "+") {
        stop("Argument 'strand' must be specified as '-' or '+'")
    }
    
    tirMatches <- initialisePackMatches()
    
    # get tir matches
    for (i in seq_len(length(Genome))) {

        matches <- Biostrings::matchPattern(
            tirSeq,
            Genome[[i]],
            max.mismatch = mismatch,
            with.indels = TRUE,
            fixed = fixed
        )
        
        if (length(matches) > 0) {
            tirMatches <- rbind(tirMatches, data.frame(
                seqnames = names(Genome)[i],
                start = GenomicRanges::start(matches),
                end = GenomicRanges::start(matches) + 
                    GenomicRanges::width(matches) - 1,
                width = GenomicRanges::width(matches),
                strand = strand
            ))
        }
    }
    
    # remove matches whose TSD sequences do not exist (index range)
    if (strand == "-") {
        removeMatch <- mapply(function(end, seqnames, tsdLength, Genome) {
            seq <- Genome[names(Genome) == seqnames][[1]]
            return((end + tsdLength) > length(seq))
        },
        tirMatches$end,
        tirMatches$seqnames,
        MoreArgs = list(tsdLength, Genome)
        )
        tirMatches <- tirMatches[removeMatch == FALSE, ]
    } else if (strand == "+") {
        tirMatches <- tirMatches[tirMatches$start > tsdLength, ]
    }
    
    return(tirMatches)
}

#' @title
#' Sample packFinder Output
#'
#' @description
#' A sample output from \code{\link{packSearch}} with 
#' cluster information. This dataframe is in the format produced by 
#' coercing a \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#'
#' @format
#' A dataframe of 9 obs. and 7 variables.
#'
#' @usage
#' data(packMatches)
#'
#' @details
#' Was obtained from running \code{\link{packSearch}} 
#' on the Arabidopsis thaliana chromosome 3 reference 
#' sequence, followed by clustering using 
#' \code{\link{packClust}}. Contains the following features:
#' \itemize{
#'     \item start - the predicted element's start base 
#'     sequence position.
#'     \item end - the predicted element's end base 
#'     sequence position.
#'     \item seqnames - character string referring to the 
#'     sequence name in \code{Genome} to which \code{start} 
#'     and \code{end} refer to.
#' }
#' 
#' The dataset was generated as in the example below.
#'
#' @examples
#' data(arabidopsisThalianaRefseq)
#' 
#' packMatches <- packSearch(
#'     Biostrings::DNAString("CACTACAA"),
#'     arabidopsisThalianaRefseq,
#'     elementLength = c(300, 3500),
#'     tsdLength = 3
#' )
#' 
#' @seealso 
#' \code{\link{packSearch}}, \code{\link{data.frame}}, 
#' \code{\link{arabidopsisThalianaRefseq}}

"packMatches"

#' @title
#' Collapse Overlapping Sequences 
#'
#' @description
#' The sequences predicted by \code{\link{packSearch}} often 
#' overlap, which may be due to the presence of closely 
#' interspersed elements or false TIR identification. 
#' In such cases, these elements can be combined using 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' in order to collapse overlapping elements, preventing 
#' over-estimation of transposon numbers. Also removes 
#' duplicate elements that have been generated in the 
#' case of multiple searches. 
#'
#' @param packMatches
#' A dataframe containing genomic ranges and names referring 
#' to sequences to be extracted. This dataframe is in the format 
#' produced by coercing a 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' 
#' Must contain the following features:
#' \itemize{
#'     \item start - the predicted element's start base 
#'     sequence position.
#'     \item end - the predicted element's end base 
#'     sequence position.
#'     \item seqnames - character string referring to the 
#'     sequence name in \code{Genome} to which \code{start} 
#'     and \code{end} refer to.
#' }
#'
#' @param Genome
#' A DNAStringSet object containing sequences referred to 
#' in \code{packMatches} (the object originally used to 
#' predict the transposons \code{\link{packSearch}}).
#' 
#' @return
#' A set of non-overlapping transposon sequences in the format
#' of the input dataframe. 
#'     
#' @seealso 
#' \code{\link{packSearch}},
#' \code{link[GenomicRanges:GRanges-class]{GRanges}}
#' 
#' @examples
#' data(packMatches)
#' data(arabidopsisThalianaRefseq)
#' 
#' packMatches$start <- 1
#' packMatches$end <- 10
#' 
#' collapseSeqs(packMatches, arabidopsisThalianaRefseq)
#' 
#' @author
#' Jack Gisby
#'
#' @export

collapseSeqs <- function(packMatches, Genome) {
    uniqueMatches <- unique(packMatches[,seq_len(5)])
    collapsedMatches <- uniqueMatches[0,]
    
    uniqueRanges <- packsToGRanges(uniqueMatches)
    collapsedRanges <- packsToGRanges(collapsedMatches)
    
    while (length(uniqueRanges) > 0) {
        query <- uniqueRanges[1]
        uniqueRanges[1] <- NULL
        
        overlaps <- GenomicRanges::findOverlaps(query, uniqueRanges, 
                                                ignore.strand = TRUE)
        
        if (length(overlaps) == 0) {
            collapsedRanges <- c(collapsedRanges, query)
        } else {
            queryStart <- GenomicRanges::start(query)
            queryEnd <- queryStart + GenomicRanges::width(query) - 1
            overlap <- uniqueRanges[S4Vectors::to(overlaps)[1]]
            overlapStart <- GenomicRanges::start(overlap)
            overlapEnd <- overlapStart + GenomicRanges::width(overlap) - 1
            
            IRanges::ranges(uniqueRanges[S4Vectors::to(overlaps)[1]]) <- 
                IRanges::IRanges(start = min(queryStart, overlapStart), 
                        end = max(queryEnd, overlapEnd))
        }
    }
    
    return(getPacksFromGRanges(collapsedRanges))
}

context("collapse")

data(packMatches)
data(arabidopsisThalianaRefseq)

packMatches$start <- 1
packMatches$end <- 10

collapsed <- collapseSeqs(packMatches, arabidopsisThalianaRefseq)

test_that("matches are collapsed", {
    expect_equal(collapsed$seqnames, as.factor("Chr3"))
    expect_equal(collapsed$start, 1)
    expect_equal(collapsed$end, 10)
    expect_equal(collapsed$width, 10)
    expect_equal(nrow(collapsed), 1)
})

unlink("tempTestOutput", recursive = TRUE)

#' @title 
#' Arabidopsis thaliana Refseq Genome Chromosome 3 Subset
#'
#' @description
#' The chromosome 3 reference sequence for Arabidopsis 
#' thaliana as a 
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}}. 
#' Can be used as a test data set, as in the associated 
#' introduction vignette. The DNA sequence between bases 
#' 10,500,000 and 14,300,000 was extracted for use in this dataset.
#'
#' @format
#' A \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#' object containing a 
#' \code{\link[Biostrings:XString-class]{DNAString}} for 
#' Arabidopsis thaliana's chromosome 3 sequence.
#'
#' @usage
#' data(arabidopsisThalianaRefseq)
#' 
#' @source
#' The Arabidopsis thaliana genome was downloaded from the 
#' NCBI refseq database on 20/SEP/2019, using 
#' \code{\link[biomartr]{getGenome}}, and chromosome 3 was 
#' extracted. The genome may also be accessed from the NCBI 
#' ftp server: \url{ftp://ftp.ncbi.nlm.nih.gov/genomes}.
#' 
#' @seealso 
#' \code{\link[biomartr]{getGenome}},
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
#' \code{\link[Biostrings:XString-class]{DNAString}},
#' \code{\link{packSearch}}
#'
#' @examples
#' data(arabidopsisThalianaRefseq)
#'
#' packMatches <- packSearch(
#'     Biostrings::DNAString("CACTACAA"),
#'     arabidopsisThalianaRefseq,
#'     elementLength = c(300, 3500),
#'     tsdLength = 3
#' )
#' 
#' @author 
#' Jack Gisby

"arabidopsisThalianaRefseq"

#' @title
#' Retrieve Saved packFinder Results (.csv)
#'
#' @description
#' Retrieves a dataframe of potential Pack-TYPE elements, 
#' previously saved using \code{\link{packSearch}} followed 
#' by \code{\link{packsToCsv}}.
#'
#' @param file
#' File path to predicted transposons in CSV format.
#'
#' @return
#' Dataframe in the format used by \code{\link{packSearch}}.
#'
#' @seealso 
#' \code{\link{packsToCsv}}, \code{\link[utils]{read.table}}, 
#' \code{\link{packSearch}}
#'
#' @examples
#' data(packMatches)
#' 
#' packMatches <- getPacksFromCsv(
#'     system.file("extdata", "packMatches.csv", package = "packFinder")
#' )
#' 
#' @author
#' Jack Gisby
#' 
#' @export

getPacksFromCsv <- function(file) {
    checkPermissions(file)
    
    return(utils::read.csv(file, stringsAsFactors = FALSE))
}

context("convert")

dir.create("tempTestOutput")
data("arabidopsisThalianaRefseq")
print(arabidopsisThalianaRefseq)
data("packMatches")

packsToCsv(packMatches, file = "tempTestOutput/packMatches.csv")

packsToFasta(packMatches = packMatches, 
            file = "tempTestOutput/packMatches.fasta", 
            Genome = arabidopsisThalianaRefseq)

packsGRanges <- packsToGRanges(packMatches)

packsFromCsv <- getPacksFromCsv(file = "tempTestOutput/packMatches.csv")
packsFromFasta <- getPacksFromFasta("tempTestOutput/packMatches.fasta")
packsFromGRanges <- getPacksFromGRanges(packsGRanges)

# ignore differences in column type
packsFromCsv$seqnames <- as.factor(packsFromCsv$seqnames)
packsFromFasta$seqnames <- as.factor(packsFromFasta$seqnames)
packsFromFasta$start <- as.integer(packsFromFasta$start)
packsFromFasta$end <- as.integer(packsFromFasta$end)
packsFromFasta$width <- as.integer(packsFromFasta$width)
packsFromGRanges$strand <- as.character(packsFromGRanges$strand)

test_that("Conversion functions create correct output from sample data", {
    expect_equal(as.character(packsFromCsv), as.character(packMatches))
    expect_equal(packsFromFasta, subset(packMatches, select = -c(cluster)))
    expect_equal(packsFromGRanges, packMatches)
})

unlink("tempTestOutput", recursive = TRUE)

#' @title
#' BLAST Analysis of PackTYPE Elements 
#'
#' @description
#' Run BLAST against user-specified databases of 
#' non-transposon and transposon-related proteins.
#' Can be used to classify transposons based on 
#' their internal sequences. 
#'
#' @param packMatches
#' A dataframe of potential Pack-TYPE transposable elements, 
#' in the format given by \code{\link{packSearch}}. This 
#' dataframe is in the format produced by coercing a 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' Will be saved as a FASTA file for VSEARCH.
#'
#' @param Genome
#' A DNAStringSet object containing sequences referred to 
#' in \code{packMatches} (the object originally used to 
#' predict the transposons \code{\link{packSearch}}). 
#' 
#' @param blastPath
#' Path to the BLAST+ executable, or name of 
#' the BLAST+ application for Linux/MacOS users.
#' 
#' @param protDb
#' For assigning Pack-TYPE elements. 
#' Path to the blast database containing nucleotide or protein
#' sequences to be matched against internal transposon 
#' sequences. Can be generated 
#' using BLAST+, or with 
#' \code{link{makeBlastDb}}.
#' 
#' @param autoDb
#' For assigning autonomous elements. 
#' Path to the blast database containing nucleotide or protein
#' sequences to be matched against internal transposon 
#' sequences. Can be generated 
#' using BLAST+, or with 
#' \code{link{makeBlastDb}}.
#' 
#' @param minE
#' Blast results with e values greater than
#' the specified cutoff will be ignored.
#' 
#' @param blastTask
#' Type of BLAST+ task, defaults to "blastn-short".
#' 
#' @param maxHits
#' Maximum hits returned by BLAST+ per query.
#' 
#' @param threads
#' Allowable number of threads to be utilised by BLAST+.
#' 
#' @param saveFolder
#' Directory to save BLAST+ results in; defaults 
#' to the working directory.
#' 
#' @param tirCutoff
#' How many bases to ignore at the terminal ends of the 
#' transposons to prevent hits to TIR sequences.
#' 
#' @return 
#' No return value; executes BLAST+ to generate hits 
#' which are stored in a .blast file in the chosen 
#' directory.
#'     
#' @seealso 
#' \code{\link{blastAnnotate}},
#' \code{\link{readBlast}}, \code{\link{packBlast}}
#' 
#' @examples
#' \dontrun{
#' packMatches <- data(packMatches)
#' Genome <- data(arabidopsisThalianaRefseq)
#' 
#' blastAnalysis(packMatches, Genome, 
#'     protDb = "C:/data/TAIR10_CDS", 
#'     autoDb = "C:/data/TAIR10_transposons", 
#'     blastPath = "C:/blast/bin/blastn.exe")
#' }
#' 
#' @references 
#' For further information, see the NCBI BLAST+ application
#' documentation and help pages 
#' (https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
#'  
#' @author
#' Jack Gisby
#'
#' @export

blastAnalysis <- function(packMatches, Genome, blastPath, protDb = NULL, 
                          autoDb = NULL, minE = 1e-3, 
                          blastTask = "blastn-short", maxHits = 100, 
                          threads = 1, saveFolder = NULL, tirCutoff = 0) {
    
    if (is.null(saveFolder)) {
        saveFolder <- getwd()
    }
    
    protDb <- file.path(getwd(), protDb)
    autoDb <- file.path(getwd(), autoDb)
    
    packMatches$id <- rownames(packMatches)
    alternatePacksToFasta(packMatches, 
                          file.path(saveFolder, "packMatches.fasta"), 
                          Genome, tirCutoff)
    fastaFile <- file.path(saveFolder, "packMatches.fasta")
    
    if (!is.null(autoDb)) {
        executeBlast("auto", fastaFile, autoDb,
                     minE, maxHits, threads, blastPath, saveFolder, blastTask)
    }
    
    if (!is.null(protDb)) {
        executeBlast("prot", fastaFile, protDb,
                     minE, maxHits, threads, blastPath, saveFolder, blastTask)
    }
}

executeBlast <- function(type, fastaFile, blastDb, minE, maxHits, 
                         threads, blastPath, saveFolder, blastTask) {
    
    system2(blastPath, args = paste0(
        "-db ", blastDb, " -query ", fastaFile,
        " -out ", file.path(saveFolder, paste0(type, "Hits.blast")),
        " -evalue ", minE,
        " -max_target_seqs ", maxHits," -num_threads ", threads, 
        " -task ", blastTask, " -word_size 7", " -outfmt 6"
    ))
}

alternatePacksToFasta <- function(packMatches, file, Genome, i) {
    
    file.create(file)
    
    # create package-specific FASTA file
    for (match in seq_len(length(packMatches[, 1]))) {
        seq <- Genome[names(Genome) == packMatches$seqnames[match]][[1]]
        seq <- seq[(packMatches$start[match] + i):(packMatches$end[match] - i)]
        write(c(
            paste0(
                ">", packMatches$id[match],
                " | seq = ", packMatches$seqnames[match],
                " | start = ", packMatches$start[match],
                " | end = ", packMatches$end[match],
                " | width = ", packMatches$width[match],
                " | strand = ", packMatches$strand[match],
                " | TSD = ", packMatches$TSD[match]
            ),
            as.character(seq)
        ),
        file = file,
        append = TRUE
        )
    }
}

#' @title
#' Global Alignment with VSEARCH
#'
#' @description
#' A global pairwise alignment of pack-TYPE elements by 
#' sequence similarity. mIt may be useful to run 
#' \code{\link{packClust}} to identify groups of similar 
#' transposable elements, before generating alignments of 
#' each group.
#'
#' @param packMatches
#' A dataframe of potential Pack-TYPE transposable elements, 
#' in the format given by \code{\link{packSearch}}. This 
#' dataframe is in the format produced by coercing a 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' Will be saved as a FASTA file for VSEARCH.
#'
#' @param Genome
#' A DNAStringSet object containing sequences referred to in 
#' \code{packMatches} (the object originally used to predict 
#' the transposons \code{\link{packSearch}}).
#'
#' @param identity
#' The sequence identity of two transposable elements in 
#' \code{packMatches} required to be grouped into a cluster.
#'
#' @param threads
#' The number of threads to be used by VSEARCH.
#'
#' @param identityDefinition 
#' The pairwise identity definition used by VSEARCH. 
#' Defaults to 2, the standard VSEARCH definition.
#'
#' @param saveFolder
#' The folder to save saveFolder files (uc, blast6out, FASTA)
#'
#' @param vSearchPath
#' When the package is run on windows systems, the 
#' location of the VSEARCH executable file must be 
#' given; this should be left as default on 
#' Linux/MacOS systems.
#'
#' @param maxWildcards
#' The maximal allowable proportion of wildcards in the 
#' sequence of each match (defaults to \code{0.05}).
#'
#' @return
#' Saves alignment information, including a \code{uc}, 
#' \code{blast6out} and a pairwise alignment \code{fasta} 
#' file, to the specified location. Returns the uc summary 
#' file generated by the alignment.
#' 
#' @note
#' In order to align sequences using VSEARCH, the 
#' executable file must first be installed.
#'
#' @seealso
#' \code{\link{tirClust}}, \code{\link{packClust}}, 
#' \code{\link{readBlast}}, \code{\link{readUc}},
#' \code{\link{filterWildcards}}, \code{\link{packSearch}}
#' 
#' 
#' @references
#' VSEARCH may be downloaded from 
#' \url{https://github.com/torognes/vsearch}, along with a 
#' manual documenting the program's parameters. See 
#' \url{https://www.ncbi.nlm.nih.gov/pubmed/27781170} for 
#' further information.
#' 
#' @examples 
#' data(arabidopsisThalianaRefseq)
#' data(packMatches)
#' 
#' # packAlign run on a Linux/MacOS system
#' \dontrun{
#'     packAlign(packMatches, Genome)
#' }
#' 
#' # packAlign run on a Windows system
#' \dontrun{
#'     packAlign(packMatches, Genome, 
#'             vSearchPath = "path/to/vsearch/vsearch.exe")
#' }
#' 
#' @author
#' Jack Gisby
#' 
#' @export

packAlign <- function(packMatches, Genome, identity = 0, threads = 1, 
                    identityDefinition = 2, maxWildcards = 0.05, saveFolder,
                    vSearchPath = "vsearch") {
    
    if (is.null(saveFolder)) {
        message("Results will be saved in the working directory: ", getwd())
        saveFolder <- getwd()
    }
    
    clustTest(saveFolder, threads, identity, strand = NULL, vSearchPath, 
            identityDefinition, type = "packAlign")

    packMatches <- 
        filterWildcards(packMatches, Genome, maxWildcards = maxWildcards)

    packMatchesFile <- file.path(saveFolder, "packMatches.fasta")
    packMatchesSet <- getPackSeqs(packMatches, Genome, output = "DNAStringSet")
    names(packMatchesSet) <- as.character(rownames(packMatches))
    Biostrings::writeXStringSet(packMatchesSet, packMatchesFile)

    system2(command = vSearchPath, args = paste0(
        "--allpairs_global ", packMatchesFile, " \ ", "--id ", identity, " \ ", 
        "--output_no_hits \ ", "--iddef ", identityDefinition, " \ ", 
        "--threads ", threads, " \ ",
        "--uc ", file.path(saveFolder, "vSearchPairwiseAlignment.uc"), " \ ",
        "--fastapairs ", file.path(saveFolder, "alignmentPairs.fasta"), " \ ",
        "--blast6out ", file.path(saveFolder, 
                                "vSearchPairwiseAlignment.blast6out"), " \ ")
    )

    return(readUc(file.path(saveFolder, "vSearchPairwiseAlignment.uc"), 
                            output = "alignment"))
}

#' @title 
#' Analyse TIR Sequences of Pre-clustered Transposable Elements
#'
#' @description
#' Takes transposable elements clustered by VSEARCH, 
#' \code{\link{packClust}}, and produces consensus sequences 
#' for the terminal inverted repeats of each. Allows for the 
#' visualisation of TIR similarities between clusters for both 
#' forward and reverse strands.
#'
#' @param packMatches
#' A dataframe containing genomic ranges and names referring 
#' to sequences to be extracted. This dataframe is in the format 
#' produced by coercing a 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' 
#' Must contain the following features:
#' \itemize{
#'     \item start - the predicted element's start base 
#'     sequence position.
#'     \item end - the predicted element's end base 
#'     sequence position.
#'     \item seqnames - character string referring to the 
#'     sequence name in \code{Genome} to which \code{start} 
#'     and \code{end} refer to.
#' }
#'
#' @param Genome
#' A DNAStringSet object containing sequences referred to in 
#' \code{packMatches} (the object originally used to predict 
#' the transposons \code{\link{packSearch}}).
#'
#' @param plot
#' Argument specifying whether the TIR consensus sequences 
#' should be plottted as a dendrogram.
#'
#' @param plotSavePath
#' File path for the dendrogram plot. If unspecified, the 
#' dendrogram plot is not saved.
#'
#' @param k
#' The k-mer size to be used for calculating a distance 
#' matrix between TIR consensus sequences. See 
#' \code{\link[kmer]{kdistance}}. Larger word sizes will not 
#' be suitable for longer TIR sequences, due to processing 
#' time required. Additionally, k must be greater than the 
#' TIR sequence length.
#'
#' @param tirLength
#' The TIR size to be considered. Consensus sequences will 
#' be generated based on the first and last \code{tirLength} 
#' bases of a transposon.
#'
#' @param output
#' Controls the output of \code{tirClust}. If output is 
#' specified as "consensus", the consensus sequences of each 
#' TIR cluster will be returned; else, if output is 
#' specified as "dendrogram", a dendrogram object will be 
#' returned for creation of customisable plots.
#'
#' @return
#' If \code{output} is specified as "consensus" (default), 
#' returns a list of consensus sequences for each cluster 
#' specified in \code{packMatches} as a 
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}}. 
#' Else if \code{output} is specified as "dendrogram", 
#' returns a dendrogram object used to create hierarchical 
#' clustering diagrams.
#'
#' @seealso
#' code{\link{packClust}}, code{\link{packAlign}},
#' \code{\link[kmer]{kdistance}},
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
#' \code{\link[ape]{as.alignment}}, \code{\link{packSearch}}
#'     
#' @author
#' Jack Gisby
#' 
#' @examples
#' data(arabidopsisThalianaRefseq)
#' data(packMatches)
#' 
#' tirClust(packMatches, arabidopsisThalianaRefseq)
#'
#' @export

tirClust <- function(packMatches, Genome, tirLength = 25, plot = TRUE,
                        plotSavePath = NULL, k = 5, output = "consensus") {
    
    if (output != "consensus" & output != "dendrogram") {
        stop("Argument 'output' must be specified 
            as 'consensus' or 'dendrogram'")
    }
    
    # use biostrings to get consensus sequences for each cluster's TIRs
    consensusSeqs <- getConsensusSeqs(packMatches, tirLength, Genome)
    
    dend <- stats::as.dendrogram(stats::hclust(
        kmer::kdistance(ape::as.DNAbin(consensusSeqs), k = k)))
    
    
    if (plot == TRUE) {
        plot(dend, main = "Clustered Transposon TIR Relationships")
    }
    
    if (!is.null(plotSavePath)) {
        grDevices::png(plotSavePath, width = 1500, height = 1000)
        plot(dend, main = "Clustered Transposon TIR Relationships")
        grDevices::dev.off()
    }
    
    if (output == "dendrogram") {
        return(dend)
    } else {
        return(consensusSeqs)
    }
}

getConsensusSeqs <- function(packMatches, tirLength, Genome) {
    fConsensusSeqs <- vector("list", 
                            length = length(unique(packMatches$clustID)))
    rConsensusSeqs <- vector("list", 
                            length = length(unique(packMatches$clustID)))
    
    for (c in seq_len(length(unique(packMatches$cluster)))) {
        clustID <- unique(packMatches$cluster)[c]
        clust <- packMatches[packMatches$cluster == clustID, ]
        forwardTirs <- vector("list", length = length(clust[, 1]))
        reverseTirs <- vector("list", length = length(clust[, 1]))
        
        for (i in seq_len(length(clust[, 1]))) {
            seq <- Genome[names(Genome) == clust$seqnames[i]][[1]]
            if (clust$strand[i] == "+") {
                fSeq <- seq[clust$start[i]:(clust$start[i] + tirLength)]
                rSeq <- seq[(clust$end[i] - tirLength):clust$end[i]]
                forwardTirs[[i]] <- fSeq
                reverseTirs[[i]] <- Biostrings::reverseComplement(rSeq)
            } else if (clust$strand[i] == "-") {
                fSeq <- seq[clust$start[i]:(clust$start[i] + tirLength)]
                rSeq <- seq[(clust$end[i] - tirLength):clust$end[i]]
                forwardTirs[[i]] <- Biostrings::reverseComplement(fSeq)
                reverseTirs[[i]] <- rSeq
            }
        }
        
        fConsensusSeqs[[c]] <- Biostrings::consensusString(
            Biostrings::DNAStringSet(forwardTirs))
        rConsensusSeqs[[c]] <- Biostrings::consensusString(
            Biostrings::DNAStringSet(reverseTirs))
    }
    
    fConsensusSeqs <- Biostrings::DNAStringSet(unlist(fConsensusSeqs))
    rConsensusSeqs <- Biostrings::DNAStringSet(unlist(rConsensusSeqs))
    
    names(fConsensusSeqs) <- 
        paste0("f", as.character(unique(packMatches$cluster)))
    names(rConsensusSeqs) <- 
        paste0("r", as.character(unique(packMatches$cluster)))
    
    return(c(fConsensusSeqs, rConsensusSeqs))
}

context("clustering functions")

dir.create("tempTestOutput")
data("arabidopsisThalianaRefseq")
data("packMatches")
load("data-r/consensusSeqs.rda")

consensusSeqTest <- tirClust(packMatches,
    arabidopsisThalianaRefseq,
    tirLength = 25,
    plot = FALSE
)

test_that("no errors thrown by tirClust", {
    expect_silent(tirClust(packMatches,
        arabidopsisThalianaRefseq,
        tirLength = 25,
        plot = FALSE,
        plotSavePath = "tempTestOutput/tirRelationships.png"
    ))
})

test_that("consensus sequences returned are as expected", {
    expect_equal(as.character(consensusSeqTest), as.character(consensusSeqs))
})

unlink("tempTestOutput", recursive = TRUE)

#' @title
#' Pack Element Filtering
#'
#' @description
#' Primary filtering stage for the \code{packSearch} 
#' algorithm. Identifies potential Pack-TYPE transposable 
#' elements based on proximity of matching inverted repeats 
#' and equality of TSD sequences.
#'
#' @param forwardMatches
#' A dataframe containing genomic ranges and names referring 
#' to forwards-facing TIR sequences and their respective 
#' TSD sequences.
#'
#' @param reverseMatches 
#' A dataframe containing genomic ranges and names referring 
#' to reverse-facing TIR sequences and their respective 
#' TSD sequences.
#'
#' @param Genome 
#' A DNAStringSet object containing the matches referred to 
#' in \code{forwardMatches} and \code{reverseMatches}
#'
#' @param elementLength
#' A vector of two integers containing the minimum and 
#' maximum transposable element length.
#' 
#' @param tsdMismatch
#' An integer referring to the allowable mismatch 
#' (substitutions or indels) between a transposon's TSD
#' sequences. \code{\link[Biostrings]{matchPattern}} from Biostrings 
#' is used for pattern matching. 
#'
#' @details
#' Used by \code{\link{packSearch}} as a primariy filtering 
#' stage. Identifies matches likely to be transposons based 
#' on their TIR region, from \code{\link{identifyTirMatches}}, 
#' and their TSD region, from \code{\link{getTsds}}. It is 
#' recommended to use the general pipeline function 
#' \code{\link{packSearch}} for identification of potential 
#' pack elements, however each stage may be called 
#' individually. Note that only exact TSD matches are 
#' considered, so supplying long sequences for TSD elements 
#' may lead to false-negative results.
#'
#' @return
#' A dataframe, \code{packMatches}, containing the locations 
#' of potential Pack-TYPE transposable elements in \code{Genome}.
#' 
#' @seealso 
#' \code{packSearch}
#' 
#' @examples 
#' data(arabidopsisThalianaRefseq)
#' 
#' forwardMatches <- identifyTirMatches(
#'     Biostrings::DNAString("CACTACAA"),
#'     arabidopsisThalianaRefseq,
#'     tsdLength = 3,
#'     strand = "+"
#' )
#' 
#' reverseMatches <- identifyTirMatches(
#'     Biostrings::reverseComplement(Biostrings::DNAString("CACTACAA")),
#'     arabidopsisThalianaRefseq,
#'     tsdLength = 3,
#'     strand = "-"
#' )
#' 
#' packMatches <- identifyPotentialPackElements(
#'     forwardMatches, 
#'     reverseMatches,
#'     arabidopsisThalianaRefseq, 
#'     c(300, 3500)
#' )
#' 
#' @author
#' Jack Gisby
#' 
#' @export

identifyPotentialPackElements <- function(forwardMatches, reverseMatches, 
                                          Genome, elementLength, 
                                          tsdMismatch = 0) {
    packMatches_list <- list()
    
    # for each forward match, consider/filter each nearby reverse match
    for (forwardMatch in seq_len(length(forwardMatches[, 1]))) {

        forwardRepeat <- forwardMatches[forwardMatch, ]
        chr <- as.character(forwardRepeat[[1]])
        searchRange <- c(forwardRepeat$start + elementLength[1], 
                         forwardRepeat$start + elementLength[2])
        
        if (searchRange[2] > length(Genome[names(Genome) == chr][[1]])) {
            searchRange[2] <- length(Genome[names(Genome) == chr][[1]])
        }
        
        reverseRepeats <- filterTsdMatches(reverseMatches, forwardRepeat, 
                                           tsdMismatch, searchRange)
        
        if (length(reverseRepeats[, 1]) > 0) {
            for (reverseMatch in seq_len(length(reverseRepeats[, 1]))) {
                
                packMatches_list[[forwardMatch]] <- data.frame(
                    seqnames = forwardRepeat$seqnames,
                    start = forwardRepeat$start,
                    end = reverseRepeats[reverseMatch, ]$end,
                    width = reverseRepeats[reverseMatch, ]$end
                    - forwardRepeat$start + 1,
                    strand = "*"
                )
            }
        }
    }
    
    packMatches <- do.call("rbind", packMatches_list)
    return(packMatches)
}

filterTsdMatches <- function(reverseMatches, forwardRepeat, tsdMismatch, 
                                searchRange) {
    
    # filters for exact matches (faster), or uses matchpattern
    if (tsdMismatch == 0) {
        reverseRepeats <- reverseMatches[
            reverseMatches$seqnames == as.character(forwardRepeat$seqnames) &
                reverseMatches$end > searchRange[1] &
                reverseMatches$end < searchRange[2] &
                reverseMatches$strand == "-" &
                reverseMatches$TSD == as.character(forwardRepeat$TSD),
            ]
    } else {
        reverseRepeats <- reverseMatches[
                reverseMatches$seqnames == as.character(forwardRepeat$seqnames)
                & reverseMatches$end > searchRange[1] &
                reverseMatches$end < searchRange[2] &
                reverseMatches$strand == "-",
            ]

        if (length(reverseRepeats[, 1]) > 0) {
            reverseRepeats <- reverseRepeats[
            Biostrings::vcountPattern(as.character(forwardRepeat$TSD),
                                        reverseRepeats$TSD,
                                        max.mismatch = tsdMismatch,
                                        with.indels = TRUE) > 0,
            ]
        }
    }

    return(reverseRepeats)
}

#' @title
#' Functional Annotation of PackTYPE Elements
#'
#' @description
#' Uses hits, previously generated using blast, 
#' to annotate transposon hits. Transposons with 
#' non-redundant transposase hits are classed as 
#' autonomous ("auto"), while others are classed as
#' "other" or "pack" based on whether the element 
#' has non-redundant hits to other proteins. 
#'  
#' @param packMatches
#' A dataframe of potential Pack-TYPE transposable elements, 
#' in the format given by \code{\link{packSearch}}. This 
#' dataframe is in the format produced by coercing a 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' Will be saved as a FASTA file for VSEARCH.
#' 
#' @param protHits
#' BLAST results for non-transposon related genes or 
#' proteins (as a \code{data.frame}). Generated using 
#' \code{\link{blastAnalysis}}.
#' 
#' @param autoHits
#' BLAST results for transposon related genes or 
#' proteins (as a \code{data.frame}). Generated using 
#' \code{\link{blastAnalysis}}.
#' 
#' @return 
#' Returns the original \code{packMatches} dataframe, 
#' with the addition of a "classification" column 
#' containing one of the following values:
#' \itemize{
#'     \item auto - elements that match known 
#'     transposases or transposon-related proteins 
#'     are classified as autonomous elements
#'     \item pack - elements that match other 
#'     proteins or genic sequences may be classified 
#'     as Pack-TYPE elements
#'     \item other - elements that generate no 
#'     significant hits
#' }
#'     
#' @seealso 
#' \code{\link{blastAnalysis}},
#' \code{\link{readBlast}}, \code{\link{packBlast}}
#' 
#' @examples
#' data("packMatches")
#' 
#' # read in some protein hits
#' p <- data.frame(
#'     query_id = c(2, 3),
#'     subject_id = c("prot", "hyp")
#' )
#' 
#' # read in some autonomous hits
#' a <- data.frame(
#'     query_id = c(3, 4),
#'     subject_id = c("transposase", "mutator")
#' )
#' 
#' blastAnnotate(p, a, packMatches)
#' 
#' @references 
#' For further information, see the NCBI BLAST+ application
#' documentation and help pages 
#' (https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
#' 
#' @note 
#' Requires that the query ids in the protein and 
#' autonomous hits match the row names in packMatches.
#'  
#' @author
#' Jack Gisby
#'
#' @export

blastAnnotate <- function(protHits, autoHits, packMatches) {
    matchType <- vector("character", length=nrow(packMatches))
    
    for (rowname in seq_len(nrow(packMatches))) {
        protHit <- nrow(protHits[protHits$query_id == as.integer(rowname),])
        autoHit <- nrow(autoHits[autoHits$query_id == as.integer(rowname),])
        
        if (autoHit > 0) {
            matchType[rowname] <- "auto"
        } else if (protHit > 0) {
            matchType[rowname] <- "pack"
        } else {
            matchType[rowname] <- "other"
        }
    }
    
    packMatches$classification <- matchType
    return(packMatches)
}

#' @title
#' Cluster Transposons with VSEARCH
#'
#' @description 
#' Cluster potential pack-TYPE elements by sequence 
#' similarity. Resulting groups may be aligned with 
#' \code{\link{packAlign}}, or the clusters may be 
#' analysed with \code{\link{tirClust}}
#'
#' @param packMatches
#' A dataframe of potential Pack-TYPE transposable elements, 
#' in the format given by \code{\link{packSearch}}. This 
#' dataframe is in the format produced by coercing a 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' Will be saved as a FASTA file for VSEARCH.
#'
#' @param Genome
#' A DNAStringSet object containing sequences referred to 
#' in \code{packMatches} (the object originally used to 
#' predict the transposons \code{\link{packSearch}}).
#'
#' @param identity
#' The sequence identity of two transposable elements in 
#' \code{packMatches} required to be grouped into a cluster.
#'
#' @param threads
#' The number of threads to be used by VSEARCH.
#'
#' @param identityDefinition
#' The pairwise identity definition used by VSEARCH. 
#' Defaults to 2, the standard VSEARCH definition.
#'
#' @param strand
#' The strand direction (+, - or *) to be clustered.
#'
#' @param saveFolder
#' The folder to save output files (uc, blast6out, FASTA)
#'
#' @param vSearchPath
#' When the package is run on windows systems, the 
#' location of the VSEARCH executable file must be 
#' given; this should be left as default on 
#' Linux/MacOS systems.
#'
#' @param maxWildcards
#' The maximal allowable proportion of wildcards in the 
#' sequence of each match (defaults to \code{0.05}).
#' 
#' @return
#' Saves cluster information, including a \code{uc} and 
#' \code{blast6out} file, to the specified location. Returns 
#' the given \code{packMatches} dataframe with an additional 
#' column, \code{cluster}, containing cluster IDs.
#' 
#' @note
#' In order to cluster sequences using VSEARCH, the 
#' executable file must first be installed.
#' 
#' @seealso
#' \code{\link{tirClust}}, \code{\link{packAlign}}, 
#' \code{\link{readBlast}}, \code{\link{readUc}},
#' \code{\link{filterWildcards}}, \code{\link{packSearch}}
#'
#' @references
#' VSEARCH may be downloaded from 
#' \url{https://github.com/torognes/vsearch}. See 
#' \url{https://www.ncbi.nlm.nih.gov/pubmed/27781170} 
#' for further information.
#' 
#' @examples 
#' data(arabidopsisThalianaRefseq)
#' data(packMatches)
#' 
#' # packClust run on a Linux/MacOS system
#' \dontrun{
#'     packClust(packMatches, Genome)
#' }
#' 
#' # packClust run on a Windows system
#' \dontrun{
#'     packClust(packMatches, Genome, 
#'             vSearchPath = "path/to/vsearch/vsearch.exe")
#' }
#' 
#' @author
#' Jack Gisby
#' 
#' @export



packClust <- function(packMatches, Genome, identity = 0.6, threads = 1, 
                        identityDefinition = 2, maxWildcards = 0.05, 
                        strand = "both", saveFolder = NULL,
                        vSearchPath = "vsearch") {
    
    if (is.null(saveFolder)) {
        message("Results will be saved in the working directory: ", getwd())
        saveFolder <- getwd()
    }
    
    strand <- clustTest(saveFolder, threads, identity, strand, vSearchPath,
                        identityDefinition, type = "packClust")

    packMatches <- filterWildcards(packMatches, Genome, 
                                    maxWildcards = maxWildcards)

    packMatchesFile <- file.path(saveFolder, "packMatches.fasta")
    ID <- as.integer(rownames(packMatches))
    packMatches$ID <- ID
    packMatches <- packMatches[order(-packMatches$width), ]
    packMatchesSet <- getPackSeqs(packMatches, Genome, output = "DNAStringSet")
    names(packMatchesSet) <- as.character(rownames(packMatches))
    Biostrings::writeXStringSet(packMatchesSet, packMatchesFile)

    system2(command = vSearchPath, args = paste0(
        "--cluster_smallmem ", packMatchesFile, " \ ", "--id ", identity, 
        " \ ", "--strand ", strand, " \ ", "--iddef ", identityDefinition, 
        " \ ", "--threads ", threads, " \ ", "--qmask none \ ", 
        "--log ", file.path(saveFolder, "packMatches.log"), " \ ", 
        "--uc ", file.path(saveFolder, "packMatches.uc"), " \ ", 
        "--blast6out ", file.path(saveFolder, "packMatches.blast6out"), " \ ", 
        "--sizeout --minseqlength 10")
    )

    vSearchClusts <- readUc(file = file.path(saveFolder, "packMatches.uc"))
    vSearchClusts <- vSearchClusts[vSearchClusts$type != "C", ]

    packMatches$strand <- mapply(function(strand) {
            if (strand == "*") { return("+")
            } else {return(strand)}
        },
        strand = as.character(vSearchClusts$strand)
    )

    packMatches$cluster <- vSearchClusts$cluster
    rownames(packMatches) <- packMatches$ID
    packMatches <- packMatches[order(packMatches$ID), ]
    return(subset(packMatches, select = -c(ID)))
}

clustTest <- function(saveFolder, threads, identity, strand, vSearchPath,
                        identityDefinition, type) {
    if (parallel::detectCores() < threads) {
        stop("There are not ", threads, " cores available")
    }
    
    if (identity > 1 | identity < 0) {
        stop("Identity must be of type integer or double, 
            and have a value between 0 and 1")
    }
    
    if (type != "packAlign") {
        if (strand == "+") {
            strand <- "plus"
        } else if (strand == "*") {
            strand <- "both"
        } else if (strand != "both" & strand != "plus") {
            message("Argument 'strand' must be specified as 'plus' or 'both'")
        }
    }
    if (!is.double(identityDefinition)) {
        stop("Argument 'identityDefinition' must 
            be an integer between 0 and 4.")
    } else if (identityDefinition > 4 | identityDefinition < 0) {
        stop("Argument 'identityDefinition' must 
            be an integer between 0 and 4.")
    }
    
    if (system2(vSearchPath, "--v") != 0) {
        message("VSEARCH cannot be found at", vSearchPath, ". Please ensure 
                that the VSEARCH executable file is installed at the correct 
                location and that the full file path is specified.")
    }
    
    if (type == "packClust") {
        return(strand)
    }
}

#' @title
#' Remove Low Quality Sequences
#'
#' @description
#' Takes transposable elements detected by 
#' \code{\link{packSearch}} and removes those with large 
#' numbers of wildcard ("N") bases. Used by 
#' \code{\link{packClust}} and \code{\link{packAlign}} to 
#' remove poor quality sequences that may interfere with 
#' the quality of sequence alignments.
#'
#' @param packMatches 
#' A dataframe containing genomic ranges 
#' and names referring to sequences to be extracted.
#'
#' @param Genome
#' The original set of sequences used to generate the 
#' transposons detected by \code{\link{packSearch}}.
#'
#' @param maxWildcards
#' The maximal allowable proportion of wildcards in the 
#' sequence of each match (defaults to \code{0.05}).
#' 
#' @return
#' The original dataframe, \code{packMatches}, with 
#' sequences removed that are found to contain a proportion 
#' of wildcards ("N") greater than that specified in 
#' \code{maxWildcards}.
#' 
#' @author
#' Jack Gisby
#' 
#' @examples
#' data(arabidopsisThalianaRefseq)
#' data(packMatches)
#'     
#' filteredMatches <- filterWildcards(
#'     packMatches, 
#'     arabidopsisThalianaRefseq, 
#'     maxWildcards = 0.05
#' )
#' 
#' @seealso 
#' \code{\link{packClust}}, \code{\link{packAlign}}, \code{packMatches},
#' data(arabidopsisThalianaRefseq)
#' 
#' @author 
#' Jack Gisby
#' 
#' @export

filterWildcards <- function(packMatches, Genome, maxWildcards = 0.05) {
    badMatches <- vector(mode = "logical", length = nrow(packMatches))
    
    message(paste0("Removing matches with a proportion of 
                    wildcards ('N's) above ", maxWildcards))
    
    # gets each sequence and removes those with many "N" bases
    for (i in seq_len(nrow(packMatches))) {
        seq <- Genome[names(Genome) == packMatches$seqnames[i]]
        seq <- as.character(seq[[1]][packMatches$start[i]:packMatches$end[i]])
        
        if (grepl("N", seq)) {
            if ((nchar(gsub("N", "", seq)) / nchar(seq)) > maxWildcards) {
                badMatches[i] <- TRUE
            }
        }
    }

    message(paste0(sum(badMatches), " matches removed"))
    return(packMatches[badMatches == FALSE, ])
}

#' @title
#' Extract Sequences of Pack-TYPE Elements
#'
#' @description
#' Method to quickly extract the sequences of predicted 
#' Pack-TYPE elements (as created by \code{\link{packSearch}}).
#'
#' @param packMatches
#' A dataframe containing genomic ranges and names referring 
#' to sequences to be extracted. This dataframe is in the format 
#' produced by coercing a 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' 
#' Must contain the following features:
#' \itemize{
#'     \item start - the predicted element's start base 
#'     sequence position.
#'     \item end - the predicted element's end base 
#'     sequence position.
#'     \item seqnames - character string referring to the 
#'     sequence name in \code{Genome} to which \code{start} 
#'     and \code{end} refer to.
#' }
#'
#' @param Genome
#' A DNAStringSet object containing sequences referred to 
#' in \code{packMatches} (the object originally used to 
#' predict the transposons \code{\link{packSearch}}).
#'
#' @param output
#' The type of object to be returned:
#' \itemize{
#'     \item output = "DNAStringSet", returns a 
#'     \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#'     object (default).
#'     \item output = "character", returns a 
#'     \code{character} vector.
#' }
#' 
#' @return
#' transposon sequences extracted from \code{packMatches}. 
#' At default returns the sequences as a 
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#' or, if \code{output} is set to "character", returns a 
#' character vector. 
#'     
#' @seealso 
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
#' \code{\link{packSearch}},
#' \code{\link[Biostrings:XString-class]{DNAString}}
#' 
#' @examples
#' data(arabidopsisThalianaRefseq)
#'
#' packMatches <- packSearch(
#'     Biostrings::DNAString("CACTACAA"),
#'     arabidopsisThalianaRefseq,
#'     elementLength = c(300, 3500),
#'     tsdLength = 3
#' )
#'
#' packSeqs <- getPackSeqs(packMatches, arabidopsisThalianaRefseq)
#'
#' @author
#' Jack Gisby
#'
#' @export

getPackSeqs <- function(packMatches, Genome, output = "DNAStringSet") {
    if (output != "string" & output != "DNAStringSet") {
        stop("Argument 'output' must be specified 
            as 'string' or 'DNAStringSet'")
    }

    # gets each sequence in packMatches
    seqs <- mapply(function(start, end, seqnames, Genome) {
            seq <- Genome[names(Genome) == seqnames][[1]][start:end]
            return(as.character(seq))
        },
        packMatches$start,
        packMatches$end,
        packMatches$seqnames,
        MoreArgs = list(Genome = Genome)
    )

    if (output == "string") {
        return(seqs)
    } else if (output == "DNAStringSet") {
        seqs <- Biostrings::DNAStringSet(seqs)
        names(seqs) <- as.character(packMatches$seqnames)
        return(seqs)
    }
}

#' @title
#' Save packFinder Results in FASTA Format (.fasta)
#'
#' @description
#' Saves a dataframe of potential Pack-TYPE elements, 
#' usually generated via \code{\link{packSearch}}. 
#' May be retrieved using \code{\link{getPacksFromFasta}}.
#'
#' @param file
#' FASTA file save path.
#'
#' @param packMatches
#' taframe containing genomic ranges and names referring 
#' to sequences to be extracted. Can be obtained from 
#' \code{\link{packSearch}} or generated from a 
#' \code{\link[GenomicRanges:GRanges-class]{GRanges}} object, 
#' after conversion to a dataframe. Must contain the 
#' following features:
#' \itemize{
#'     \item start - the predicted element's start base 
#'     sequence position.
#'     \item end - the predicted element's end 
#'     base sequence position.
#'     \item seqnames - character string 
#'     referring to the sequence name in \code{Genome} to 
#'     which \code{start} and \code{end} refer to.
#' }
#'
#' @param Genome
#' A DNAStringSet object containing sequences referred to 
#' in \code{packMatches} (the object originally used to 
#' predict the transposons \code{\link{packSearch}}).
#'
#' @return
#' Save location of Fasta file.
#'
#' @seealso 
#' \code{\link{getPacksFromFasta}}, \code{\link{packSearch}}
#' 
#' @examples
#' data(arabidopsisThalianaRefseq)
#' data(packMatches)
#' 
#' packsToFasta(
#'     packMatches, 
#'     system.file("extdata", "packMatches.fasta", package = "packFinder"), 
#'     arabidopsisThalianaRefseq
#' )
#' 
#' @author
#' Jack Gisby
#'
#' @export

packsToFasta <- function(packMatches, file, Genome) {
    
    file.create(file)
    
    # create package-specific FASTA file
    for (match in seq_len(length(packMatches[, 1]))) {
        seq <- Genome[names(Genome) == packMatches$seqnames[match]][[1]]
        seq <- seq[packMatches$start[match]:packMatches$end[match]]
        write(c(
            paste0(
                ">", packMatches$seqnames[match],
                " | start = ", packMatches$start[match],
                " | end = ", packMatches$end[match],
                " | width = ", packMatches$width[match],
                " | strand = ", packMatches$strand[match],
                " | TSD = ", packMatches$TSD[match]
            ),
            as.character(seq)
        ),
        file = file,
        append = TRUE
        )
    }
    
    return(print(paste0("FASTA written to ", file)))
}

#' @title
#' packFinder Algorithm Pipeline
#'
#' @description
#' General use pipeline function for the Pack-TYPE transposon
#' finding algorithm.
#'
#' @param tirSeq
#' A \code{\link[Biostrings:DNAString-class]{DNAString}} 
#' object containing the TIR sequence to be searched for.
#'
#' @param Genome
#' A \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#' object to be searched.
#'
#' @param mismatch
#' The maximum edit distance to be considered for TIR 
#' matches (indels + substitions). See 
#' \code{\link[Biostrings]{matchPattern}} for details.
#'
#' @param elementLength
#' The maximum element length to be considered, as a vector 
#' of two integers. E.g. \code{c(300, 3500)}
#'
#' @param tsdLength
#' Integer referring to the length of the flanking TSD region.
#' 
#' @param fixed
#' Logical that will be passed to the `fixed` argument of 
#' \code{\link[Biostrings]{matchPattern}}. Determines the behaviour of IUPAC
#' ambiguity codes when searching for TIR sequences.
#' 
#' @param tsdMismatch
#' An integer referring to the allowable mismatch 
#' (substitutions or indels) between a transposon's TSD
#' sequences. \code{\link[Biostrings]{matchPattern}} from Biostrings 
#' is used for pattern matching. 
#'
#' @details
#' Finds potential pack-TYPE elements based on:
#' \itemize{
#'     \item Similarity of TIR sequence to \code{tirSeq}
#'     \item Proximity of potential TIR sequences
#'     \item Directionality of TIR sequences
#'     \item Similarity of TSD sequences
#' }
#'
#' The algorithm finds potential forward and reverse TIR 
#' sequences using \code{\link{identifyTirMatches}} and 
#' their associated TSD sequence via \code{\link{getTsds}}. 
#' The main filtering stage, 
#' \code{\link{identifyPotentialPackElements}}, filters 
#' matches to obtain a dataframe of potential PACK elements. 
#' Note that this pipeline does not consider the 
#' possibility of discovered elements being autonomous 
#' elements, so it is recommended to cluster and/or BLAST 
#' elements for further analysis. Furthermore, only exact TSD 
#' matches are considered, so supplying long sequences for 
#' TSD elements may lead to false-negative results.
#'
#' @return
#' A dataframe, containing elements 
#' identified by thealgorithm. These may be autonomous or 
#' pack-TYPE elements. Will contain the following features:
#' \itemize{
#'     \item start - the predicted element's start base 
#'     sequence position.
#'     \item end - the predicted element's end base 
#'     sequence position.
#'     \item seqnames - character string referring to the 
#'     sequence name in \code{Genome} to which \code{start} 
#'     and \code{end} refer to.
#'     \item width - the width of the predicted element.
#'     \item strand - the strand direction of the 
#'     transposable element. This will be set to "*" as the 
#'     \code{packSearch} function does not consider 
#'     transposons to have a direction - only TIR sequences. 
#'     Passing the \code{packMatches} dataframe to 
#'     \code{\link{packClust}} will assign a direction to 
#'     each predicted Pack-TYPE element.
#' }
#' 
#' This dataframe is in the format produced by 
#' coercing a \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. Downstream 
#' functions, such as \code{\link{packClust}}, use this 
#' dataframe to manipulate predicted transposable elements.
#'
#' @note
#' This algorithm does not consider:
#' \itemize{
#'     \item Autonomous elements - autonomous elements will 
#'     be predicted by this algorithm as there is no BLAST 
#'     step. It is recommended that, after clustering 
#'     elements using \code{\link{packClust}}, the user 
#'     analyses each group to determine which predicted 
#'     elements are autonomous and which are likely 
#'     Pack-TYPE elements. Alternatively, databases such as 
#'     Repbase (\url{https://www.girinst.org/repbase/}) 
#'     supply annotations for autonomous transposable 
#'     elements that can be used to filter autonomous matches.
#'     \item TSD Mismatches - if two TIRs do not have exact 
#'     matches for their terminal site duplications they 
#'     will be ignored. Supplying longer TSD sequences will 
#'     likely lead to a lower false-positive rate, however 
#'     may also cause a greater rate of false-negative results.
#' }
#' 
#' Pattern matching is done via \code{\link[Biostrings]{matchPattern}}.
#'
#' @seealso
#' \code{\link{identifyTirMatches}}, \code{\link{getTsds}},
#' \code{\link{identifyPotentialPackElements}}, \code{\link{packClust}},
#' \code{\link{packMatches}}, 
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
#' \code{\link[Biostrings:DNAString-class]{DNAString}},
#' \code{\link[Biostrings]{matchPattern}}
#' 
#' 
#' @examples 
#' data(arabidopsisThalianaRefseq)
#'
#' packMatches <- packSearch(
#'     Biostrings::DNAString("CACTACAA"),
#'     arabidopsisThalianaRefseq,
#'     elementLength = c(300, 3500),
#'     tsdLength = 3
#' )
#' 
#' @author
#' Jack Gisby
#' 
#' @export

packSearch <- function(tirSeq, Genome, mismatch = 0, elementLength, tsdLength, 
                       tsdMismatch = 0, fixed = TRUE) {
    
    searchCheck(mismatch, tsdLength, elementLength, tirSeq, Genome)

    message("Getting forward matches")
    forwardMatches <- identifyTirMatches(tirSeq = tirSeq, Genome = Genome,
        mismatch = mismatch, strand = "+", tsdLength = tsdLength, fixed = fixed)
    
    forwardMatches$TSD <- getTsds(tirMatches = forwardMatches, Genome = Genome,
        tsdLength = tsdLength, strand = "+")
    message(length(forwardMatches[, 1]), " forward matches identified.")
    
    message("Getting reverse matches")
    reverseMatches <- identifyTirMatches(
        tirSeq = Biostrings::reverseComplement(tirSeq), Genome = Genome, 
        mismatch = mismatch, strand = "-", tsdLength = tsdLength, fixed = fixed)
    
    reverseMatches$TSD <- getTsds(tirMatches = reverseMatches, Genome = Genome,
        tsdLength = tsdLength, strand = "-")
    message(length(reverseMatches[, 1]), " reverse matches identified.")
    
    if (length(forwardMatches[, 1]) == 0 | length(reverseMatches[, 1]) == 0) {
        message("No matches identified")
        return(NULL)
    }

    message("Filtering matches based on TSD sequences")
    packMatches <- identifyPotentialPackElements(
        forwardMatches = forwardMatches, reverseMatches = reverseMatches,
        Genome = Genome, elementLength = elementLength, 
        tsdMismatch = tsdMismatch
    )
    
    packMatches$TSD <- getTsds(tirMatches = packMatches, Genome = Genome,
                                tsdLength = tsdLength, strand = "+")
    
    message("Initial filtering complete. ", length(packMatches[, 1]), 
            " elements predicted.")
    return(packMatches)
}

searchCheck <- function(mismatch, tsdLength, elementLength, tirSeq, Genome) {
    if (!is.numeric(mismatch) | !is.numeric(tsdLength)) {
        stop("Arguments 'mismatch' and 'tsdLength' must be integers")
    }
    if (!is.vector(elementLength) | length(elementLength) != 2) {
        stop("Argument 'elementLength' must be a vector of minimum 
            and maximum transposon lengths")
    }
    if (!is.numeric(elementLength[1]) | !is.numeric(elementLength[2])) {
        stop("Vector 'elementLength' must contain integers")
    }
    if (!methods::is(tirSeq, "DNAString")) {
        stop("Argument 'tirSeq' must be of type 
            Biostrings::DNAString or character")
    }
    if (!methods::is(Genome, "DNAStringSet")) {
        stop("Argument 'Genome' must be of type Biostrings::DNAStringSet.
            You may convert files using Biostrings::readDNAStringSet
            or convert objects using Biostrings::DNAStringSet")
    }
}

context("packSearch")

dir.create("tempTestOutput")
data("arabidopsisThalianaRefseq")
data("packMatches")

packMatchesTest <- packSearch(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    mismatch = 0,
    elementLength = c(300, 3500),
    tsdLength = 3
)

test_that("dimensions of packMatches are correct", {
    expect_equal(nrow(packMatchesTest), 6)
    expect_equal(ncol(packMatchesTest), 6)

    expect_equal(nrow(packMatches), 6)
    expect_equal(ncol(packMatches), 7)
})

test_that("packSearch returns correct results", {
    packMatchesTest$strand <- "+"
    packMatchesTest$seqnames <- as.factor(packMatchesTest$seqnames)
    expect_equal(packMatchesTest, subset(packMatches, select = -c(cluster)))
})

unlink("tempTestOutput", recursive = TRUE)

#' @title
#' Retrieve Saved packFinder Results (.fasta)
#'
#' @description
#' Retrieves a dataframe of potential Pack-TYPE elements, 
#' previously saved using \code{\link{packSearch}} followed 
#' by \code{\link{packsToFasta}}. Parses the .fasta file 
#' and title field containing:
#' \itemize{
#'     \item seqnames - name of origin sequence
#'     \item start - transposon base start position on 
#'     origin sequence
#'     \item end - transposon base end position on origin 
#'     sequence
#'     \item width - width of transposon
#'     \item strand - direction of transposon 
#'     ("+", "-" or "*")
#'     \item TSD - terminal site duplication (TSD) sequence
#' }
#'
#' @param file
#' Path to predicted transposons in FASTA format.
#'
#' @return
#' Dataframe in the format used by \code{\link{packSearch}}.
#'
#' @seealso
#' \code{\link{packsToFasta}}, \code{\link{packSearch}}
#' 
#' @examples
#' data(arabidopsisThalianaRefseq)
#' data(packMatches)
#' 
#' packMatches <- getPacksFromFasta(
#'     system.file("extdata", "packMatches.fasta", package = "packFinder")
#' )
#' 
#' @author
#' Jack Gisby
#'
#' @export

getPacksFromFasta <- function(file) {
    checkPermissions(file)
    fileCon <- file(file, "r")
    
    packMatches <- initialisePackMatches(TSD = TRUE)
    
    while (TRUE) {
        seqName <- readLines(fileCon, n = 1)
        
        # read FASTA until end of file
        if ((length(seqName) == 0) | (length(seq) == 0)) {
            break
        } else if (substr(seqName, 1, 1) != ">") {
            break
        }
        seq <- readLines(fileCon, n = 1)

        # read package-specific FASTA file into dataframe
        seqName <- gsub(">", "", seqName)
        seqName <- gsub("start =", "", seqName)
        seqName <- gsub("end =", "", seqName)
        seqName <- gsub("width =", "", seqName)
        seqName <- gsub("strand =", "", seqName)
        seqName <- gsub("TSD =", "", seqName)
        seqName <- gsub(" ", "", seqName)
        seqName <- strsplit(seqName, "|", fixed = TRUE)[[1]]
        names(seqName) <- c("seqnames", "start", "end", 
                            "width", "strand", "TSD")
        
        packMatches <- rbind(packMatches, seqName, stringsAsFactors = FALSE)
        seqName <- NULL
    }
    
    close(fileCon)
    colnames(packMatches) <- c("seqnames", "start", "end", 
                                "width", "strand", "TSD")
    return(packMatches)
}

checkPermissions <- function(file) {
    if (!is.null(file)) {
        # check for read/write permissions
        if (!(file.access(file, 2) == 0)) {
            stop("File does not exist, or R does not 
                have read permissions")
        }
    }
}

initialisePackMatches <- function(TSD = FALSE) {
    if (TSD == TRUE) {
        packMatches <- data.frame(
            seqnames = character(),
            start = integer(),
            end = integer(),
            width = integer(),
            strand = character(),
            TSD = character()
        )
    } else {
        packMatches <- data.frame(
            seqnames = character(),
            start = integer(),
            end = integer(),
            width = integer(),
            strand = character()
        )
    }
    
    return(packMatches)
}

#' @title
#' Save packFinder Results in CSV Format (.csv)
#'
#' @description
#' Saves a dataframe of potential Pack-TYPE elements, 
#' usually generated via \code{\link{packSearch}}. May be 
#' retrieved using \code{\link{getPacksFromCsv}}.
#'
#' @param file
#' CSV file save path.
#'
#' @param packMatches
#' A dataframe containing genomic ranges and names referring 
#' to sequences to be extracted. Can be obtained from 
#' \code{\link{packSearch}} or generated from a 
#' \code{\link[GenomicRanges:GRanges-class]{GRanges}} object, 
#' after conversion to a dataframe. Must contain the 
#' following features:
#' \itemize{
#'     \item start - the predicted element's start base 
#'     sequence position.
#'     \item end - the predicted element's end base 
#'     sequence position.
#'     \item seqnames - character string referring to 
#'     the sequence name in \code{Genome} to which 
#'     \code{start} and \code{end} refer to.
#' }
#'
#' @return
#' Save location of csv file.
#'
#' @seealso 
#' \code{\link{getPacksFromCsv}}, \code{\link[utils]{write.table}},
#' \code{\link{packSearch}}
#' 
#' @examples
#' data(packMatches)
#' 
#' packsToCsv(
#'     packMatches, 
#'     system.file("extdata", "packMatches.csv", package = "packFinder")
#' )
#' 
#' @author
#' Jack Gisby
#'
#' @export

packsToCsv <- function(packMatches, file) {
    utils::write.csv(packMatches, file, row.names = FALSE)

    return(print(paste0("File successfully written to ", file)))
}

# this script was used to generate the file found at:
# inst/extdata/packMatches.uc

library(packFinder)

data("arabidopsisThalianaRefseq")
data("packMatches")

packClusts <- packClust(
    packMatches, 
    arabidopsisThalianaRefseq,
    saveFolder = system.file("extdata", package = "packFinder")
)

#' @title
#' Get Flanking Terminal Site Duplication Sequences
#'
#' @description
#' Gets the flanking TSD sequences of TIRs or predicted 
#' Pack-TYPE transposable elements. A dataframe of these 
#' elements can be in \code{tirMatches}.
#'
#' @param tirMatches
#' A dataframe containing genomic ranges and names referring 
#' to TIR sequences or predicted Pack-TYPE transposable 
#' elements. Should be in the format used by 
#' \code{\link{packSearch}}.
#'
#' @param Genome
#' A \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#' object containing sequences referred to in \code{tirMatches}.
#'
#' @param tsdLength 
#' The length of the TSD region to be retrieved (integer).
#'
#' @param strand
#' The strand of the TIR; "+" for forward, "-" for reverse. 
#' If the TSD sequences of transposable elements are being 
#' predicted, then this parameter can be left as default 
#' ("+"); if the TSD sequences of TIRs are being found then 
#' the strand direction must be supplied.
#'
#' @param output
#' The type of object to be returned:
#' \itemize{
#'     \item output = "DNAStringSet", returns a 
#'     \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#'     object.
#'     \item output = "character", returns a 
#'     \code{character} vector (default).
#' }
#'
#' @details
#' Called by \code{\link{packSearch}}. It is recommended 
#' to use the general pipeline function 
#' \code{\link{packSearch}} for identification of potential 
#' pack elements, which returns TSD sequences as a feature 
#' of results, however each stage may be called individually.
#'
#' @return
#' Flanking TSD sequences as a vector of characters, or if 
#' output is specified as "DNAStringSet", TSD sequences 
#' will be returned as a 
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
#' object.
#' 
#' @seealso 
#' \code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
#' \code{\link{packSearch}}, \code{tirMatches}
#' 
#' @examples 
#' data(arabidopsisThalianaRefseq)
#' data(packMatches)
#' 
#' tsdSeqs <- getTsds(packMatches, arabidopsisThalianaRefseq, 3)
#'
#' @author
#' Jack Gisby
#'
#' @export

getTsds <- function(tirMatches, Genome, tsdLength, strand = "+",
                        output = "character") {
    
    if (strand != "+" & strand != "-") {
        stop("Argument 'strand' must be specified as a character, '+' or '-'")
    }

    if (output != "character" & output != "DNAStringSet") {
        stop("Argument 'output' must be specified 
            as 'string' or 'DNAStringSet'")
    }

    # get TSD sequence in forward or reverse direction
    if (strand == "-") {
        TSDs <- mapply(function(seqnames, end, tsdLength, Genome) {
                seq <- Genome[names(Genome) == seqnames][[1]]
                return(as.character(seq[(end + 1):(end + tsdLength)]))
            },
            tirMatches$seqnames,
            tirMatches$end,
            MoreArgs = list(tsdLength = tsdLength, Genome = Genome)
        )
    } else if (strand == "+") {
        TSDs <- mapply(function(seqnames, start, tsdLength, Genome) {
                seq <- Genome[names(Genome) == seqnames][[1]]
                return(as.character(seq[(start - tsdLength):(start - 1)]))
            },
            tirMatches$seqnames,
            tirMatches$start,
            MoreArgs = list(tsdLength = tsdLength, Genome = Genome)
        )
    }
    
    if (output == "DNAStringSet") {
        return(Biostrings::DNAStringSet(TSDs))
    } else {
        return(TSDs)
    }
}

#' @title 
#' packFinder: a package for the de novo Annotation 
#' of Pack-TYPE Transposable Elements
#'
#' @description
#' Algorithm and tools for in silico pack-TYPE transposon 
#' discovery. Filters a given genome for properties unique 
#' to DNA transposons and provides tools for the 
#' investigation of returned matches. 
#' 
#' @section Main Algorithm:
#' The goal of packFinder was to implement a simple 
#' tool for the prediction of potential Pack-TYPE 
#' elements. packFinder uses the following prior 
#' knowledge, provided by the user, to detect transposons:
#' 
#' \itemize{
#'     \item Terminal Inverted Repeat (TIR) Base Sequence
#'     \item Length of Terminal Site Duplication (TSD)
#'     \item Length of the Transposon
#' }
#' 
#' These features provide enough information to detect 
#' autonomous and pack-TYPE elements. For a transposon 
#' to be predicted by packFinder its TSD sequences 
#' must be identical to each other, its forward TIR 
#' sequence must match the base sequence provided and 
#' its reverse TIR sequence must match its reverse complement.
#' 
#' Transposons are therefore predicted by searching a 
#' given genome for these characteristics, and further 
#' analysis steps can reveal the nature of these elements 
#' - while the packFinder tool is sensitive for the 
#' detection of transposons, it does not discriminate 
#' between autonomous and Pack-TYPE elements. Autonomous 
#' elements will contain a transposase gene within the 
#' terminal inverted repeats and tend to be larger than 
#' their Pack-TYPE counterparts; pack-TYPE elements 
#' instead capture sections of host genomes. Following 
#' cluster analysis, BLAST can be used to discern 
#' which predicted elements are autonomous 
#' (transposase-containing) and with are true 
#' Pack-TYPE elements.
#' 
#' @section Workflow:
#' An example of a standard workflow can be found using 
#' \code{browseVignettes(package = "packFinder")}. The 
#' primary functions include:
#' 
#' \itemize{
#'     \item packSearch - the packSearch algorithm 
#'     uses simple pattern matching to detect 
#'     DNA transposons.
#'     \item packClust - VSEARCH is used for 
#'     clustering elements based on sequence 
#'     similarity. 
#' }
#' 
#' Having obtained the sequences of transposable
#' elements in a given genome, it is recommende 
#' to carry out a BLAST search for each transposon 
#' cluster. This can identify which elements are 
#' likely autonomous, and which may be  Pack-TYPE.
#' 
#' The \code{packFinder} functions report the 
#' position of elements in a given genome using a
#' dataframe in the format of \code{\link{packMatches}}. 
#' This dataframe is in the format produced by 
#' coercing a \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' object to a dataframe: \code{data.frame(GRanges)}. 
#' 
#' @name 
#' packFinder
#' 
#' @seealso 
#' \code{\link{packSearch}}
#'
#' @author 
#' Jack Gisby

NULL

#' @title
#' Export packFinder Results to a GRanges Object
#'
#' @description
#' A dataframe containing genomic ranges and names referring 
#' to sequences to be extracted, likely obtained from 
#' \code{\link{packSearch}}, can be converted to a GRanges 
#' object. Can be converted back to a dataframe using 
#' \code{\link{getPacksFromGRanges}}. Additional features, 
#' such as clusters and TSD sequences, will be included in 
#' the object as metadata columns.
#'
#'
#' @param packMatches
#' A dataframe containing genomic ranges and names 
#' referring to sequences to be extracted. Can be obtained 
#' from \code{\link{packSearch}} or generated from a 
#' \code{\link[GenomicRanges:GRanges-class]{GRanges}} object, 
#' after conversion to a dataframe. Must contain the 
#' following features:
#' \itemize{
#'     \item start - the predicted element's start 
#'     base sequence position.
#'     \item end - the predicted element's end base 
#'     sequence position.
#'     \item seqnames - character string referring 
#'     to the sequence name in \code{Genome} to which 
#'     \code{start} and \code{end} refer to.
#' }
#'
#' @return
#' A GRanges object containing the ranges contained in 
#' \code{packMatches} and additional metadata columns. May 
#' be easily converted between dataframe and GRanges format 
#' for use in the \code{packFinder} package and 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}} 
#' package. Note that most functions in the \code{packFinder} 
#' package require sequence ranges to be provided in 
#' dataframe format.
#'
#' @seealso 
#' \code{\link{getPacksFromGRanges}},
#' \code{link[GenomicRanges:GRanges-class]{GRanges}}
#' 
#' @examples 
#' data(packMatches)
#' packGRanges <- packsToGRanges(packMatches)
#' 
#' @author
#' Jack Gisby
#'
#' @export

packsToGRanges <- function(packMatches) {
    return(GenomicRanges::GRanges(packMatches))
}

#' @title
#' Convert NCBI BLAST+ Files to Dataframe
#'
#' @description
#' Reads .blast6out files (NCBI Blast Format) generated by 
#' the VSEARCH clustering and alignment algorithms.
#'
#' @param file
#' The file path of the blast file.
#' 
#' @param minE
#' Blast results with e values greater than
#' the specified cutoff will be ignored.
#' 
#' @param length
#' Blast results alignment lengths lower below
#' this value will be ignored
#' 
#' @param identity
#' Blast results with target sequence identities below
#' this value will be ignored.
#' 
#' @param removeExactMatches
#' If true, matches with 100% sequence identity will 
#' be ignored to prevent self-hits. 
#' 
#' @param scope
#' If specified, blast results below the specified value
#' will be ignored. Note that the dataframe of transposon
#' matches must also be supplied to calculate scope. Scope is 
#' the proportion of the transposon's internal sequence 
#' occupied by the BLAST hit. 
#' 
#' @param packMatches
#' taframe containing genomic ranges and names referring 
#' to sequences to be extracted. Can be obtained from 
#' \code{\link{packSearch}} or generated from a 
#' \code{\link[GenomicRanges:GRanges-class]{GRanges}} object, 
#' after conversion to a dataframe. Must contain the 
#' following features:
#' \itemize{
#'     \item start - the predicted element's start base 
#'     sequence position.
#'     \item end - the predicted element's end 
#'     base sequence position.
#'     \item seqnames - character string 
#'     referring to the sequence name in \code{Genome} to 
#'     which \code{start} and \code{end} refer to.
#' }
#'
#' @return
#' A dataframe containing the converted .blast6out file. 
#' The file contains the following features:
#' \itemize{
#'     \item Query sequence ID
#'     \item Target sequence ID
#'     \item Percenty sequence identity
#'     \item Alignment length
#'     \item Number of mismatches
#'     \item Number of gaps
#'     \item Base position of alignment start 
#'     in query sequence
#'     \item Base position of alignment end in query sequence
#'     \item Base position of alignment start in target sequence
#'     \item Base position of alignment end in target sequence
#'     \item E-value
#'     \item Bit score
#' }
#'
#' @details
#' blast6out file is tab-separated text file compatible with 
#' NCBI BLAST m8 and NCBI BLAST+ outfmt 6 formats. One 
#' cluster/alignment can be found for each line.
#'
#' @seealso
#' code{\link{blastAnalysis}}, code{\link{blastAnnotate}},
#' code{\link{packAlign}}, 
#' code{\link{readUc}}, code{\link{packClust}}
#'
#' @references
#' For further information, see the NCBI BLAST+ application
#' documentation and help pages 
#' (https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
#' VSEARCH may be downloaded from 
#' \url{https://github.com/torognes/vsearch}; see 
#' \url{https://www.ncbi.nlm.nih.gov/pubmed/27781170} 
#' for further information.
#'
#' @examples 
#' readBlast(system.file(
#'     "extdata", 
#'     "packMatches.blast6out", 
#'     package = "packFinder"
#' ))
#' 
#' @author
#' Jack Gisby
#'
#' @export

readBlast <- function(file, minE = 1, length = 0, identity = 0,
                      removeExactMatches = FALSE, 
                      scope = NULL, packMatches = NULL) {
    
    checkPermissions(file)
    
    blastData <- utils::read.table(file, sep = "\t", 
                           header = FALSE, stringsAsFactors = FALSE)
    
    colnames(blastData) <- c("query_id", "subject_id","identity", 
                             "alignment_length", "mismatches", "gap_opens", 
                             "q.start", "q.end", "s.start", "s.end", "evalue", 
                             "bit_score")
    
    blastData <- blastData[blastData$evalue < minE,]
    blastData <- blastData[blastData$alignment_length > length,]
    blastData <- blastData[blastData$identity > identity,]
    
    if (removeExactMatches) {
        blastData <- blastData[blastData$identity < 100,]
    }
    
    if (!is.null(scope)) {
        width <- vector("integer", nrow(blastData))
        scope_filter <- vector("numeric", nrow(blastData))
        
        for (rowname in seq_len(nrow(blastData))) {
            width[rowname] <- 
                packMatches$width[
                    packMatches$id == blastData[rowname,]$query_id]
            scope_filter[rowname] <- 
                blastData[rowname,]$alignment_length / width[rowname]
        }
        
        blastData$width <- width
        blastData$scope_filter <- scope_filter
        
        blastData <- blastData[blastData$scope_filter > scope,]
    }
    
    return(blastData)
}

context("conversion functions")

dir.create("tempTestOutput")
ucClust <- readUc("data-raw/clustOutput.uc")
ucAlign <- readUc("data-raw/alignOutput.uc")
blast6Out <- readBlast("data-raw/clustOutput.blast6out")

test_that("object type is as expected", {
    expect_type(ucClust, "list")
    expect_type(ucAlign, "list")
    expect_type(blast6Out, "list")
})

test_that("object dimensions are as expected", {
    expect_equal(nrow(ucClust), 11)
    expect_equal(ncol(ucClust), 8)

    expect_equal(nrow(ucAlign), 16)
    expect_equal(ncol(ucAlign), 8)

    expect_equal(nrow(blast6Out), 1)
    expect_equal(ncol(blast6Out), 12)
})

test_that("object output is as expected", {
    expect_equal(max(ucClust$cluster), 4)
})

unlink("tempTestOutput", recursive = TRUE)

#' @title
#' Make Blast Database 
#'
#' @description
#' Generates a BLAST database to be queried. Required 
#' for identifying sequences using the BLAST+ 
#' software. 
#'
#' @param fastaFile
#' FASTA file containing sequences to generate a BLAST 
#' database from. 
#' 
#' @param dbPath
#' Path to save the BLAST database to. 
#' 
#' @param blastPath
#' Path/name of BLAST program to use. Name of
#' the application for Linux/MacOS, absolute 
#' path for the executable for windows users. 
#' 
#' @param dbType
#' Type of BLAST database to create, e.g. "nucl"
#' for a nucleotide database. 
#' 
#' @return 
#' No return value; generates a blast database in 
#' the chosen directory. 
#'     
#' @seealso 
#' \code{\link{packSearch}}
#' 
#' @examples
#' \dontrun{
#' makeBlastDb("genes.fasta", "blastdb.db", "C:/blast.exe")
#' }
#' 
#' @references 
#' For further information, see the NCBI BLAST+ application
#' documentation and help pages 
#' (https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
#'  
#' @author
#' Jack Gisby
#'
#' @export

makeBlastDb <- function(fastaFile, dbPath, blastPath, dbType = "nucl") {
    system2(blastPath, args = paste0(
        "-in ", file.path(getwd(), fastaFile), " ", 
        "-input_type fasta -dbtype ", dbType, 
        " -parse_seqids -out ", dbPath, " ",
        "-title packfinder_DB"   
    ))
}

#' @title
#' Convert .uc Files to Dataframe
#'
#' @description
#' Reads .uc files (USEARCH Cluster Format) generated by 
#' the VSEARCH clustering and alignment algorithms.
#'
#' @param file
#' The file path of the .uc file.
#'
#' @param output
#' The type of analysis that was carried out to produce the 
#' .uc file.
#' \itemize{
#'     \item If output is specified as "cluster", VSEARCH 
#'     clustering was carried out.
#'     \item If output is specified as "alignment", 
#'     VSEARCH pairwise global alignment was carried out.
#' }
#' 
#' Note that clustering produces one "H" record 
#' for each sequence, and one "C" record for 
#' each cluster, while an alignment produces an 
#' "H" record for each alignment (see details).
#'
#' @return
#' A dataframe containing the converted .uc file. The fields 
#' contained within are as follows:
#' \itemize{
#'     \item Record type - "H, C or N", see details for 
#'     further information.
#'     \item Cluster designation 
#'     (\code{output = "cluster"} only)
#'     \item Sequence length, or cluster size
#'     \item Percent identity to target
#'     \item The nucleotide strand 
#'     (\code{output = "cluster"} only)
#'     \item A compressed alignment - see details for 
#'     further information.
#'     \item ID of query sequence
#'     \item ID of target sequence ("H" records only)
#' }
#'
#' @details
#' USEARCH cluster format is a tab separated text file that 
#' contains clustering and/or alignment information for a 
#' set of sequences. For each sequence a record type, 
#' "H, C or N", is provided providing information about the 
#' type of "hit" in the dataframe. These refer to:
#' 
#' \itemize{
#'     \item H - Hit - for alignments, indicates an 
#'     identified alignment of two supplied sequences. For 
#'     clustering, indicates the cluster assignment for a query.
#'     \item C - Cluster record - a record for each cluster generated.
#'     \item N - No hit - indicates that no cluster was 
#'     assigned or no alignment was found with a target 
#'     sequence. For clustering, a query with no hits becomes 
#'     the centroid of a new cluster.
#' }
#' 
#' Additionally, for each record a "compressed alignment" 
#' is generated. This is the alignment represented in a 
#' compact format including the letters "M", "D", and "I". 
#' Before each letter, the number of consecutive columns 
#' of the given letter type is also given. The letter types 
#' are as follows:
#' 
#' \itemize{
#'     \item "M" - Match - Identical bases between the query 
#'     and target sequence
#'     \item "D" - Deletion - A gap in the target sequence
#'     \item "I" - Insertion - A gap in the query sequence
#' }
#' 
#' An example of this would be "13M", referring to 13 
#' consecutive matches between the query and target sequence.
#' 
#' @seealso
#' code{\link{tirClust}}, code{\link{packAlign}}, 
#' code{\link{readBlast}}, code{\link{packClust}}
#'
#' @references
#' VSEARCH may be downloaded from 
#' \url{https://github.com/torognes/vsearch}. See 
#' \url{https://www.ncbi.nlm.nih.gov/pubmed/27781170} 
#' for further information.
#'
#' @examples 
#' readUc(system.file(
#'     "extdata", 
#'     "packMatches.uc", 
#'     package = "packFinder"
#' ))
#' 
#' @author
#' Jack Gisby
#' 
#' @export

readUc <- function(file, output = "cluster") {
    checkPermissions(file)

    if (output != "cluster" & output != "alignment") {
        stop("Argument 'output' must be specified as 'cluster' or 'alignment'")
    }

    packClusts <- utils::read.table(file, sep = "\t")
    colnames(packClusts) <- c(
        "type",
        "cluster",
        "width",
        "identity",
        "strand",
        "6",
        "7",
        "cigarAlignment",
        "query",
        "target"
    )

    cluster <- packClusts$cluster
    strand <- packClusts$strand

    # different parts of the uc file are required for different purposes
    if (output == "cluster") {
        return(subset(packClusts, select = -c(6, 7)))
    } else if (output == "alignment") {
        return(subset(packClusts, select = -c(cluster, strand, 6, 7)))
    }
}

#' @title
#' Retrieve packFinder Results from GRanges Object
#'
#' @description
#' A \code{link[GenomicRanges:GRanges-class]{GRanges}}  
#' object, potentially 
#' generated using \code{\link{packSearch}} and 
#' \code{\link{packsToGRanges}}, can be converted to a 
#' dataframe. If a GRanges object is supplied without TSD 
#' information, this can be calculated and appended to the 
#' final dataframe.
#'
#' @param packGRanges
#' \code{link[GenomicRanges:GRanges-class]{GRanges}}
#' object to be coerced.
#'
#' @param Genome
#' (optional) Sequences referred to by \code{packGRanges}.
#'
#' @param tsdLength
#' (optional) Length of TSD sequences.
#'
#' @return 
#' Dataframe in the format used by \code{\link{packSearch}}. 
#' If \code{Genome} and \code{tsdLength} are supplied, then 
#' TSD sequences are retrieved and returned 
#' as part of the dataframe.
#' 
#' @seealso
#' \code{\link{packsToGRanges}}, 
#' \code{link[GenomicRanges:GRanges-class]{GRanges}},
#' \code{\link{packSearch}} 
#' 
#' @examples
#' data(packMatches)
#' 
#' GRangesObject <- packsToGRanges(packMatches)
#' packMatches <- getPacksFromGRanges(GRangesObject)
#' 
#' @author
#' Jack Gisby
#'
#' @export

getPacksFromGRanges <- function(packGRanges, Genome = NULL, tsdLength = NULL) {
    # dataframe to GRanges, with the option to get TSDs as metadata
    if (is.null(Genome) | is.null(tsdLength)) {
        return(as.data.frame(packGRanges))
    }
    else if (!is.null(Genome) & !is.null(tsdLength)) {
        packMatches <- as.data.frame(packGRanges)
        packMatches$TSDs <- getTsds(packMatches, Genome, tsdLength)
        return(packMatches)
    }
}

context("data")

dir.create("tempTestOutput")
data("arabidopsisThalianaRefseq")
data("packMatches")

test_that("length of datasets is as expected", {
    expect_equal(length(arabidopsisThalianaRefseq), 1)
    expect_equal(sum(arabidopsisThalianaRefseq@ranges@width), 3800001)
    expect_equal(length(packMatches[, 1]), 6)
    expect_equal(length(packMatches[1, ]), 7)
})

test_that("type of dataset is as expected", {
    expect_is(arabidopsisThalianaRefseq, "DNAStringSet")
    expect_is(packMatches, "data.frame")
})

unlink("tempTestOutput", recursive = TRUE)

---
title: "packFinder"
author: "Jack Gisby"
date: "`r Sys.Date()`" 
output:
    BiocStyle::html_document:
        number_sections: yes
        toc: true
vignette: >
    %\VignetteIndexEntry{packFinder}
    %\VignetteEngine{knitr::rmarkdown}
    %\VignetteEncoding{UTF-8} 
---

```{r setup, include = FALSE}
BiocStyle::markdown()

library(packFinder)

dir.create("tempOutput")
```

# Introduction

The goal of packFinder was to implement a simple tool for the prediction of 
potential Pack-TYPE elements. packFinder uses the following prior knowledge, 
provided by the user, to detect transposons:

1. Terminal Inverted Repeat (TIR) Base Sequence
2. Length of Terminal Site Duplication (TSD)
3. Length of the Transposon

These features, shown in __Figure 1__, provide enough information to detect 
autonomous and pack-TYPE elements. For a transposon to be predicted by 
packFinder its TSD sequences must be identical to each other, its forward TIR 
sequence must match the base sequence provided and its reverse TIR sequence 
must match its reverse complement.

![**Important structural features of Pack-TYPE transposons**](tirSeq.jpg)

Transposons are therefore predicted by searching a given genome for these 
characteristics, and further analysis steps can reveal the nature of these 
elements - while the packFinder tool is sensitive for the detection of 
transposons, it does not discriminate between autonomous and Pack-TYPE 
elements. 

Autonomous elements will contain a transposase gene within the 
terminal inverted repeats and tend to be larger than their Pack-TYPE 
counterparts; pack-TYPE elements instead capture sections of host genomes. 
Following cluster analysis, BLAST can be used to discern which predicted 
elements are autonomous (transposase-containing) and with are true 
Pack-TYPE elements.

# Getting Started

Users may download packFinder and use the primary function - packSearch - 
to locate potential transposons in a given set of DNA sequences. In addition 
to R packages, the command line tool `VSEARCH` must be installed prior to use 
of clustering and alignment functions.

## R Package Dependencies

CRAN and Bioconductor packages will be automatically installed 
when downloading `packFinder`. The package may be installed 
from the development branch of Bioconductor, as long as R version 
4.0.0 is installed. 

```{r packFinderInstall, eval=FALSE}
if (!requireNamespace("BiocManager", quietly = TRUE))
    install.packages("BiocManager")

# The following initializes usage of Bioc devel
BiocManager::install(version='devel')

BiocManager::install("packFinder")
```

## Command Line Dependencies

While this allows R to run the packSearch pipeline within packFinder, 
VSEARCH must be installed for use of clustering and alignment functions 
within packSearch. Detailed installation instructions are available from 
the README file on the VSEARCH github https://github.com/torognes/vsearch). 
The command line can be used to install VSEARCH on Linux and MacOS operating 
systems (using wget and tar) while VSEARCH can be downloaded and extracted for 
use on Windows systems. 

For Linux and MacOS systems, correct installation of VSEARCH should allow 
users to use all functions within packFinder whereas for windows users, the 
absolute path to the VSEARCH executable file must be specified when calling 
packFinder clustering and alignment functions. This vignette will assume a 
MacOS/Linux operating system - see ?packClust for examples using windows.

# Searching for Potential Transposable Elements using packFinder

## Getting Data

Both DNA to be searched and the TIR query must be in XString format 
(see Biostrings). The TIR query should be coerced to a DNAString object 
while the DNA sequence, or set of sequences, should be a DNAStringSet. 
Bioconductor's `Biostrings` provides methods for the conversion of various 
formats to `DNAString` objects, including from character vectors and 
fasta files.

```{r biostrings}
# Convert a character vector to a DNAString
Biostrings::DNAString("CATG")

# Convert a list of character vectors to a DNAStringSet
Biostrings::DNAStringSet(c(
    "CATG",
    "GTAC"
))

# Convert a FASTA file to a DNAStringSet
Biostrings::readDNAStringSet(
    system.file("extdata", "packMatches.fasta", package = "packFinder"),
    format = "fasta"
)
```


## Using packSearch

In this example we search a subset of chromosome 3, from the Arabidopsis 
thaliana reference sequence; our query will be first 8 bases of the CACTA1 
autonomous transposable element's TIR. The CACTA transposons have TSD 
sequences that are 3 bases long, and the Pack-CACTA elements tend to be 
between 300 to 3500 bases in width. Since the first 8 bases of the Pack-CACTA 
TIRs tend to be conserved between elements, the default allowable mismatch of 
0 was used. 

```{r packSearch}
data("arabidopsisThalianaRefseq")

packMatches <- packSearch(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    elementLength = c(300, 3500),
    tsdLength = 3
)

head(packMatches)
```

In this subset, we identified six potential Pack-TYPE elements, however at 
this stage it is unclear what genetic information can be found between the 
inverted TIR sequences __Figure 1__. These elements could contain transposase 
genes, making them autonomous elements, or may be Pack-TYPE elements that 
have captured parts of the Arabidopsis thaliana host genome. Additionally, 
of the Pack-TYPE elements detected, some may share sequences of related 
chromosomal origin between their TIRs. 

`packSearch` returns a dataframe of ranges, in the format produced by 
coercing a GRanges object to a dataframe: `data.frame(GRanges)`. 
This format is used to transfer the transposon ranges between the functions 
in `packFinder`.

# Analysing Potential Transposable Elements using packFinder

## Clustering of Transposable Elements

In order to make downstream analysis more efficient and understand the 
relations between the identified elements, we can use VSEARCH to cluster 
predicted transposons. Here we run VSEARCH with the default parameters, 
meaning:

*An identity of 60% between two elements is required to form a cluster
*The method of identity detection is the VSEARCH default

The identity and method of identity definition can be altered depending on 
analysis (see VSEARCH documentation). 

```{r packClustData, include=FALSE}
# We don't have the dependencies for clustering, so get pre-clustered results
data(packMatches)
```
```{r packClust, eval=FALSE}
packMatches <- packClust(
    packMatches,
    arabidopsisThalianaRefseq,
    saveFolder = "tempOutput"
)
```
```
## Rognes T, Flouri T, Nichols B, Quince C, Mahe F (2016)
## VSEARCH: a versatile open source tool for metagenomics
## PeerJ 4:e2584 doi: 10.7717/peerj.2584 https://doi.org/10.7717/peerj.2584
##
## vsearch v2.14.1_win_x86_64, 7.9GB RAM, 4 cores
## https://github.com/torognes/vsearch
##
## vsearch v2.14.1_win_x86_64, 7.9GB RAM, 4 cores
## https://github.com/torognes/vsearch
##
## Reading file data/packMatches.fasta 100%
## 9396 nt in 6 seqs, min 713, max 2463, avg 1566
## Counting k-mers 100%
## Clustering 100%
## Sorting clusters 100%
## Writing clusters 100%
## Clusters: 5 Size min 1, max 2, avg 1.2
## Singletons: 4, 66.7% of seqs, 80.0% of clusters
## Sorting clusters by abundance 100%
```
```
##   seqnames   start     end width strand TSD cluster
## 1     Chr3  100830  102347  1518      + TGT       3
## 2     Chr3 1068802 1069514   713      + ATA       4
## 3     Chr3 2807747 2809454  1708      + GGT       2
## 4     Chr3 3487540 3488267   728      + ATA       4
## 5     Chr3 3582297 3584562  2266      + ATA       1
## 6     Chr3 3738747 3741209  2463      + TTG       0
```

Of the 6 elements identified in this data subset, only two were found to be 
in the same cluster. When a new cluster is created, the transposon is 
designated as being on the forwards strand (+); elements that are subsequently 
assigned to this cluster are given a strand designation relative to the 
original element in the cluster. 

For this data, as there are few clusters, 
all of the elements have been designated as being on the forwards strand. 
Adjusting the identity % required or changing the identity definition could 
lead to more effective clustering, or lead to false-positives. 

Note, by default `filterWildcards` is called when clustering 
or aligning sequences; this prevents low quality sequences from clustering 
together, and by default removes sequences with a proportion of wildcards 
("N") above 5%. 

## Reading VSEARCH Output Files

Based on the results of packClust, we found that the second and fourth matches 
have similar sequences. Additionally, these potential elements have a similar 
width and so it is feasible that these are elements that have been duplicated 
by a transposase. To investigate the extent of the similarities, we can read 
the more detailed VSEARCH output files:

* USEARCH cluster format `.uc` - containing a summary of the clustering done 
by VSEARCH
* BLAST output - a BLAST compatible summary containing details of the BLAST 
matches between clusters 


```{r readClust}
readUc(system.file(
    "extdata",
    "packMatches.uc",
    package = "packFinder"
))

readBlast(system.file(
    "extdata",
    "packMatches.blast6out",
    package = "packFinder"
))
```

## Clustering of TIR Sequences

Additionally, tirClust can provide a summary of the similarities between 
the TIR sequences of clustered transposons. While in this example "CACTACAA" 
has been used as the TIR search query, the CACTA TIR sequence is longer than 
8 base pairs - although the rest of the TIR sequence may be less well 
conserved. 

For each cluster, tirClust creates a consensus sequence for 
the forward and reverse TIR regions; in this case we will consider the first 
25 base pairs of each TIR. Additionally, clustering of these TIRs is carried 
out using kmer clustering before being plotted as a dendrogram for 
visualisation. 

```{r tirClust}
consensusSeqs <- tirClust(packMatches,
    arabidopsisThalianaRefseq,
    tirLength = 25
)

head(consensusSeqs)
```

As expected, the forward and reverse TIRs of each transposon are very similar; 
this is also true for the two clustered transposons. From the dendrogram, 
groups are visible that weren't found by the VSEARCH clustering; this indicates 
that, while the TIR sequences are related, these groups likely have different 
genetic material between their TIR sequences.

## Alignment of Transposable Elements

After clusters of transposable elements have been identified, it may be useful 
to produce an alignment. Since we know that the transposable elements 
identified by VSEARCH have a minimum 60% sequence similarity, it will be 
possible to produce good quality sequence alignments. This can be useful 
for downstream analysis, such as BLAST searching. In this instance, an 
alignment was done for cluster 4; so an alignment of only two sequences 
was carried out. 

```{r align, eval=FALSE}
packMatches <- packAlign(
    packMatches,
    arabidopsisThalianaRefseq,
    saveFolder = "tempOutput"
)
```
```
## Rognes T, Flouri T, Nichols B, Quince C, Mahe F (2016)
## VSEARCH: a versatile open source tool for metagenomics
## PeerJ 4:e2584 doi: 10.7717/peerj.2584 https://doi.org/10.7717/peerj.2584
##
## vsearch v2.14.1_win_x86_64, 7.9GB RAM, 4 cores
## https://github.com/torognes/vsearch
##
## vsearch v2.14.1_win_x86_64, 7.9GB RAM, 4 cores
## https://github.com/torognes/vsearch
##
## Reading file data//packMatches.fasta 100%
## 9396 nt in 6 seqs, min 713, max 2463, avg 1566
## Masking 100%
## Aligning 100%
## Matching query sequences: 5 of 6 (83.33%)
```

As with VSEARCH clustering, a `.uc` and `.blast` file are created by the 
alignment. The alignment function returns a summary of the `.uc` file, 
however the full details of each file can be recovered as before. 

## BLAST Analysis
BLAST search has not been implemented in this package, so it is expected that 
the user will define clusters of transposable elements above before BLAST 
investigation. This stage will identify the tranposable elements that are 
autonomous and confirm which matches are likely pack-TYPE elements. 

Since 
this example included only a small subset of the Arabidopsis thaliana genome, 
and there were very few elements identified, a BLASTN search was carried out 
for cluster 4. In a larger study, it would have been wise to instead select 
more interesting clusters and BLAST a subset of these. 

```{r blast}
# the packMatches dataframe is exported as a FASTA file for NCBI blast search
packsToFasta(
    packMatches,
    "tempOutput/packMatches.fasta",
    arabidopsisThalianaRefseq
)
```

After running the cluster 4 sequence using NCBI's BLASTX service, two major 
hits were found (GENBANK: AAD28056.1 and AAF06087.1) for the sequence. One 
match appeared to be similar to a putative En/Spm-like transposon protein, 
suggesting that this transposon may not be a Pack-TYPE element as it may not 
have captured genetic material from the host genome.

# Data Formats and Conversion

## packSearch Dataframe

Dataframes are used by packFinder to contain the locations of predicted 
elements and additional metadata, such as TSD sequence and cluster 
designation. The dataframe is in a similar format to that stored by 
GenomicRanges. The dataframe can be saved and restored, as below, for longer 
term storage and export of packFinder results. packMatches refers to the 
object created in previous steps, by applying packSearch followed by packClust.

```{r csvConvert}
packsToCsv(packMatches, "tempOutput/packMatches.csv")
print(getPacksFromCsv("tempOutput/packMatches.csv"))
```

## GRanges

As mentioned previously, the packFinder dataframe is in a similar format to 
that stored by GenomicRanges. The data produced by packFinder can therefore 
be quickly converted to and from a GRanges object, as below. 

```{r grangesConvert}
packsGRanges <- packsToGRanges(packMatches)
print(packsGRanges)

print(getPacksFromGRanges(packsGRanges))
```

If many elements are discovered by packFinder it may be necessary to identify 
overlapping elements, that could not be produced by a transposase, and remove 
them; this can be done using GenomicRanges. 

## FASTA

Additionally, the dataframe produced by packFinder can be exported to and 
restored from FASTA formats. While this will not save additional metadata 
columns, such as cluster designations, the core information will be
preserved in the FASTA title field. 

```{r fastaConvert}
packsToFasta(
    packMatches,
    "tempOutput/packMatches.fasta",
    arabidopsisThalianaRefseq
)

print(getPacksFromFasta("tempOutput/packMatches.fasta"))
```

# Step-wise packSearch Functions

While it is recommended to use packSearch to identify potential Pack-TYPE 
elements, it is possible to run the individual functions that make up the 
packSearch pipeline. Below are the steps used by packSearch to predict 
transposons, using the same parameters as the previous example. 

## Identifying Potential TIRs

Potential TIR sequences are first identified by pattern matching. 

```{r identifyTirMatches}
forwardMatches <- identifyTirMatches(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    strand = "+",
    tsdLength = 3
)
nrow(forwardMatches)

reverseMatches <- identifyTirMatches(
    Biostrings::reverseComplement(Biostrings::DNAString("CACTACAA")),
    arabidopsisThalianaRefseq,
    strand = "-",
    tsdLength = 3
)
nrow(reverseMatches)
```

## Obtaining TSD Sequences 

The function `getTsds` may be used to quickly obtain the TSD sequences for 
TIRs. This function can also be used for obtaining TSDs from the ranges of 
known full transposons, given a dataframe in the format produced by packSearch. 

```{r getTsds}
forwardMatches$TSD <- getTsds(
    forwardMatches,
    arabidopsisThalianaRefseq,
    3,
    strand = "+"
)

head(forwardMatches)

reverseMatches$TSD <- getTsds(
    reverseMatches,
    arabidopsisThalianaRefseq,
    3,
    strand = "-"
)

head(reverseMatches)
```

## Filtering TIR Matches

The main step of the algorithm matches TIR sequences together based on their 
proximity and TSD sequence. 

```{r identifyPotentialPackElements}
identifyPotentialPackElements(
    forwardMatches,
    reverseMatches,
    arabidopsisThalianaRefseq,
    c(300, 3500)
)
```

After attaching the TSD sequences to the dataframe of matches, the output 
matches that of the previous example run using the packSearch pipeline. 

## Get Transposon Sequences 

The function `getSeqs` may be used to obtain transposon sequences from a 
dataframe in the format produced by packSearch. This could additionally be 
used on other dataframes of ranges, such as that produced by 
identifyTirMatches above. 

```{r getPackSeqs}
getPackSeqs(packMatches, arabidopsisThalianaRefseq)
```

# References
Much of the information on the properties of Pack-TYPE transposable elements 
was obtained from the following papers during the development of 
`packFinder`.

* Catoni, M. et al. (2019) ‘Mobilization of Pack-CACTA transposons in 
Arabidopsis suggests the mechanism of gene shuffling’, 
*Nucleic Acids Research*. Oxford University Press, 47(3), pp. 1311–1320. 
doi: 10.1093/nar/gky1196.
* Jiang, N. et al. (2004) ‘Pack-MULE transposable elements mediate gene 
evolution in plants’, *Nature*, 431(7008), pp. 569–573. 
doi: 10.1038/nature02953.

The Bioconductor packages `GenomicRanges` and `Biostrings` were 
used extensively for detection of tranposable elements, and manipulation of 
DNA sequences. 

Finally, the `packFinder` package has now been described in *PLOS Genetics*:

Gisby, J. S. & Catoni, M. (2022) The widespread nature of Pack-TYPE transposons 
reveals their importance for plant genome evolution. *PLOS Genetics* 18(2): 
e1010078. doi: 
[10.1371/journal.pgen.1010078](https://doi.org/10.1371/journal.pgen.1010078)

# Session Information

This vignette was compiled during the following session:

```{r SessionInfo}
sessionInfo()
```

```{r include=FALSE}
unlink("tempOutput", recursive = TRUE)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readUc.R
\name{readUc}
\alias{readUc}
\title{Convert .uc Files to Dataframe}
\usage{
readUc(file, output = "cluster")
}
\arguments{
\item{file}{The file path of the .uc file.}

\item{output}{The type of analysis that was carried out to produce the 
.uc file.
\itemize{
    \item If output is specified as "cluster", VSEARCH 
    clustering was carried out.
    \item If output is specified as "alignment", 
    VSEARCH pairwise global alignment was carried out.
}

Note that clustering produces one "H" record 
for each sequence, and one "C" record for 
each cluster, while an alignment produces an 
"H" record for each alignment (see details).}
}
\value{
A dataframe containing the converted .uc file. The fields 
contained within are as follows:
\itemize{
    \item Record type - "H, C or N", see details for 
    further information.
    \item Cluster designation 
    (\code{output = "cluster"} only)
    \item Sequence length, or cluster size
    \item Percent identity to target
    \item The nucleotide strand 
    (\code{output = "cluster"} only)
    \item A compressed alignment - see details for 
    further information.
    \item ID of query sequence
    \item ID of target sequence ("H" records only)
}
}
\description{
Reads .uc files (USEARCH Cluster Format) generated by 
the VSEARCH clustering and alignment algorithms.
}
\details{
USEARCH cluster format is a tab separated text file that 
contains clustering and/or alignment information for a 
set of sequences. For each sequence a record type, 
"H, C or N", is provided providing information about the 
type of "hit" in the dataframe. These refer to:

\itemize{
    \item H - Hit - for alignments, indicates an 
    identified alignment of two supplied sequences. For 
    clustering, indicates the cluster assignment for a query.
    \item C - Cluster record - a record for each cluster generated.
    \item N - No hit - indicates that no cluster was 
    assigned or no alignment was found with a target 
    sequence. For clustering, a query with no hits becomes 
    the centroid of a new cluster.
}

Additionally, for each record a "compressed alignment" 
is generated. This is the alignment represented in a 
compact format including the letters "M", "D", and "I". 
Before each letter, the number of consecutive columns 
of the given letter type is also given. The letter types 
are as follows:

\itemize{
    \item "M" - Match - Identical bases between the query 
    and target sequence
    \item "D" - Deletion - A gap in the target sequence
    \item "I" - Insertion - A gap in the query sequence
}

An example of this would be "13M", referring to 13 
consecutive matches between the query and target sequence.
}
\examples{
readUc(system.file(
    "extdata", 
    "packMatches.uc", 
    package = "packFinder"
))

}
\references{
VSEARCH may be downloaded from 
\url{https://github.com/torognes/vsearch}. See 
\url{https://www.ncbi.nlm.nih.gov/pubmed/27781170} 
for further information.
}
\seealso{
code{\link{tirClust}}, code{\link{packAlign}}, 
code{\link{readBlast}}, code{\link{packClust}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packsToCsv.R
\name{packsToCsv}
\alias{packsToCsv}
\title{Save packFinder Results in CSV Format (.csv)}
\usage{
packsToCsv(packMatches, file)
}
\arguments{
\item{packMatches}{A dataframe containing genomic ranges and names referring 
to sequences to be extracted. Can be obtained from 
\code{\link{packSearch}} or generated from a 
\code{\link[GenomicRanges:GRanges-class]{GRanges}} object, 
after conversion to a dataframe. Must contain the 
following features:
\itemize{
    \item start - the predicted element's start base 
    sequence position.
    \item end - the predicted element's end base 
    sequence position.
    \item seqnames - character string referring to 
    the sequence name in \code{Genome} to which 
    \code{start} and \code{end} refer to.
}}

\item{file}{CSV file save path.}
}
\value{
Save location of csv file.
}
\description{
Saves a dataframe of potential Pack-TYPE elements, 
usually generated via \code{\link{packSearch}}. May be 
retrieved using \code{\link{getPacksFromCsv}}.
}
\examples{
data(packMatches)

packsToCsv(
    packMatches, 
    system.file("extdata", "packMatches.csv", package = "packFinder")
)

}
\seealso{
\code{\link{getPacksFromCsv}}, \code{\link[utils]{write.table}},
\code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packBlast.R
\name{packBlast}
\alias{packBlast}
\title{Pipeline for BLAST/Classification of PackTYPE Elements}
\usage{
packBlast(
  packMatches,
  Genome,
  blastPath,
  protDb,
  autoDb,
  minE = 0.001,
  blastTask = "blastn-short",
  maxHits = 100,
  threads = 1,
  saveFolder = NULL,
  tirCutoff = 100,
  autoCutoff = 1e-05,
  autoLength = 150,
  autoIdentity = 70,
  autoScope = NULL,
  protCutoff = 1e-05,
  protLength = 250,
  protIdentity = 70,
  protScope = 0.3
)
}
\arguments{
\item{packMatches}{A dataframe of potential Pack-TYPE transposable elements, 
in the format given by \code{\link{packSearch}}. This 
dataframe is in the format produced by coercing a 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. 
Will be saved as a FASTA file for VSEARCH.}

\item{Genome}{A DNAStringSet object containing sequences referred to 
in \code{packMatches} (the object originally used to 
predict the transposons \code{\link{packSearch}}).}

\item{blastPath}{Path to the BLAST+ executable, or name of 
the BLAST+ application for Linux/MacOS users.}

\item{protDb}{For assigning Pack-TYPE elements. 
Path to the blast database containing nucleotide or protein
sequences to be matched against internal transposon 
sequences. Can be generated 
using BLAST+, or with 
\code{link{makeBlastDb}}.}

\item{autoDb}{For assigning autonomous elements. 
Path to the blast database containing nucleotide or protein
sequences to be matched against internal transposon 
sequences. Can be generated 
using BLAST+, or with 
\code{link{makeBlastDb}}.}

\item{minE}{Blast results with e values greater than
the specified cutoff will be ignored. This will 
be passed to BLASTN and applied to both transposon 
and non-transposon matches.}

\item{blastTask}{Type of BLAST+ task, defaults to "blastn-short".}

\item{maxHits}{Maximum hits returned by BLAST+ per query.}

\item{threads}{Allowable number of threads to be utilised by BLAST+.}

\item{saveFolder}{Directory to save BLAST+ results in; defaults 
to the working directory.}

\item{tirCutoff}{How many bases to ignore at the terminal ends of the 
transposons to prevent hits to TIR sequences.}

\item{autoCutoff}{Blast results for transposon-related elements will be 
filtered to 
ignore those with e values above the specified cutoff.}

\item{autoLength}{Blast results for transposon-related elements containing 
hits with 
alignment lengths lower than this value will be ignored}

\item{autoIdentity}{Blast results for transposon-related elements containing 
hits with 
sequence identities lower than this value will be ignored}

\item{autoScope}{If specified, transposon-related blast results below the 
specified value
will be ignored. Note that the dataframe of transposon
matches must also be supplied to calculate scope. Scope is 
the proportion of the transposon's internal sequence 
occupied by the BLAST hit.}

\item{protCutoff}{Blast results for genic/other matches will be 
filtered to 
ignore those with e values above the specified cutoff.}

\item{protLength}{Blast results for genic/other matches containing 
hits with 
alignment lengths lower than this value will be ignored}

\item{protIdentity}{Blast results for genic/other matches containing 
hits with 
sequence identities lower than this value will be ignored}

\item{protScope}{If specified, genic/other blast matches below the 
specified value
will be ignored. Note that the dataframe of transposon
matches must also be supplied to calculate scope. Scope is 
the proportion of the transposon's internal sequence 
occupied by the BLAST hit.}
}
\value{
Returns the original \code{packMatches} dataframe, 
with the addition of a "classification" column 
containing one of the following values:
\itemize{
    \item auto - elements that match known 
    transposases or transposon-related proteins 
    are classified as autonomous elements
    \item pack - elements that match other 
    proteins or genic sequences may be classified 
    as Pack-TYPE elements
    \item other - elements that generate no 
    significant hits
}
}
\description{
Run BLAST against user-specified databases of 
non-transposon and transposon-relates proteins.
Can be used to classify transposons based on 
their internal sequences.
}
\examples{
\dontrun{
packMatches <- data(packMatches)
Genome <- data(arabidopsisThalianaRefseq)

packBlast(packMatches, Genome, 
    protDb = "C:/data/TAIR10_CDS", 
    autoDb = "C:/data/TAIR10_transposons", 
    blastPath = "C:/blast/bin/blastn.exe")
}

}
\references{
For further information, see the NCBI BLAST+ application
documentation and help pages 
(https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
}
\seealso{
\code{\link{blastAnalysis}}, \code{\link{packSearch}},
\code{\link{readBlast}}, \code{\link{blastAnnotate}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getPacksFromFasta.R
\name{getPacksFromFasta}
\alias{getPacksFromFasta}
\title{Retrieve Saved packFinder Results (.fasta)}
\usage{
getPacksFromFasta(file)
}
\arguments{
\item{file}{Path to predicted transposons in FASTA format.}
}
\value{
Dataframe in the format used by \code{\link{packSearch}}.
}
\description{
Retrieves a dataframe of potential Pack-TYPE elements, 
previously saved using \code{\link{packSearch}} followed 
by \code{\link{packsToFasta}}. Parses the .fasta file 
and title field containing:
\itemize{
    \item seqnames - name of origin sequence
    \item start - transposon base start position on 
    origin sequence
    \item end - transposon base end position on origin 
    sequence
    \item width - width of transposon
    \item strand - direction of transposon 
    ("+", "-" or "*")
    \item TSD - terminal site duplication (TSD) sequence
}
}
\examples{
data(arabidopsisThalianaRefseq)
data(packMatches)

packMatches <- getPacksFromFasta(
    system.file("extdata", "packMatches.fasta", package = "packFinder")
)

}
\seealso{
\code{\link{packsToFasta}}, \code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packClust.R
\name{packClust}
\alias{packClust}
\title{Cluster Transposons with VSEARCH}
\usage{
packClust(
  packMatches,
  Genome,
  identity = 0.6,
  threads = 1,
  identityDefinition = 2,
  maxWildcards = 0.05,
  strand = "both",
  saveFolder = NULL,
  vSearchPath = "vsearch"
)
}
\arguments{
\item{packMatches}{A dataframe of potential Pack-TYPE transposable elements, 
in the format given by \code{\link{packSearch}}. This 
dataframe is in the format produced by coercing a 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. 
Will be saved as a FASTA file for VSEARCH.}

\item{Genome}{A DNAStringSet object containing sequences referred to 
in \code{packMatches} (the object originally used to 
predict the transposons \code{\link{packSearch}}).}

\item{identity}{The sequence identity of two transposable elements in 
\code{packMatches} required to be grouped into a cluster.}

\item{threads}{The number of threads to be used by VSEARCH.}

\item{identityDefinition}{The pairwise identity definition used by VSEARCH. 
Defaults to 2, the standard VSEARCH definition.}

\item{maxWildcards}{The maximal allowable proportion of wildcards in the 
sequence of each match (defaults to \code{0.05}).}

\item{strand}{The strand direction (+, - or *) to be clustered.}

\item{saveFolder}{The folder to save output files (uc, blast6out, FASTA)}

\item{vSearchPath}{When the package is run on windows systems, the 
location of the VSEARCH executable file must be 
given; this should be left as default on 
Linux/MacOS systems.}
}
\value{
Saves cluster information, including a \code{uc} and 
\code{blast6out} file, to the specified location. Returns 
the given \code{packMatches} dataframe with an additional 
column, \code{cluster}, containing cluster IDs.
}
\description{
Cluster potential pack-TYPE elements by sequence 
similarity. Resulting groups may be aligned with 
\code{\link{packAlign}}, or the clusters may be 
analysed with \code{\link{tirClust}}
}
\note{
In order to cluster sequences using VSEARCH, the 
executable file must first be installed.
}
\examples{
data(arabidopsisThalianaRefseq)
data(packMatches)

# packClust run on a Linux/MacOS system
\dontrun{
    packClust(packMatches, Genome)
}

# packClust run on a Windows system
\dontrun{
    packClust(packMatches, Genome, 
            vSearchPath = "path/to/vsearch/vsearch.exe")
}

}
\references{
VSEARCH may be downloaded from 
\url{https://github.com/torognes/vsearch}. See 
\url{https://www.ncbi.nlm.nih.gov/pubmed/27781170} 
for further information.
}
\seealso{
\code{\link{tirClust}}, \code{\link{packAlign}}, 
\code{\link{readBlast}}, \code{\link{readUc}},
\code{\link{filterWildcards}}, \code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readBlast.R
\name{readBlast}
\alias{readBlast}
\title{Convert NCBI BLAST+ Files to Dataframe}
\usage{
readBlast(
  file,
  minE = 1,
  length = 0,
  identity = 0,
  removeExactMatches = FALSE,
  scope = NULL,
  packMatches = NULL
)
}
\arguments{
\item{file}{The file path of the blast file.}

\item{minE}{Blast results with e values greater than
the specified cutoff will be ignored.}

\item{length}{Blast results alignment lengths lower below
this value will be ignored}

\item{identity}{Blast results with target sequence identities below
this value will be ignored.}

\item{removeExactMatches}{If true, matches with 100% sequence identity will 
be ignored to prevent self-hits.}

\item{scope}{If specified, blast results below the specified value
will be ignored. Note that the dataframe of transposon
matches must also be supplied to calculate scope. Scope is 
the proportion of the transposon's internal sequence 
occupied by the BLAST hit.}

\item{packMatches}{taframe containing genomic ranges and names referring 
to sequences to be extracted. Can be obtained from 
\code{\link{packSearch}} or generated from a 
\code{\link[GenomicRanges:GRanges-class]{GRanges}} object, 
after conversion to a dataframe. Must contain the 
following features:
\itemize{
    \item start - the predicted element's start base 
    sequence position.
    \item end - the predicted element's end 
    base sequence position.
    \item seqnames - character string 
    referring to the sequence name in \code{Genome} to 
    which \code{start} and \code{end} refer to.
}}
}
\value{
A dataframe containing the converted .blast6out file. 
The file contains the following features:
\itemize{
    \item Query sequence ID
    \item Target sequence ID
    \item Percenty sequence identity
    \item Alignment length
    \item Number of mismatches
    \item Number of gaps
    \item Base position of alignment start 
    in query sequence
    \item Base position of alignment end in query sequence
    \item Base position of alignment start in target sequence
    \item Base position of alignment end in target sequence
    \item E-value
    \item Bit score
}
}
\description{
Reads .blast6out files (NCBI Blast Format) generated by 
the VSEARCH clustering and alignment algorithms.
}
\details{
blast6out file is tab-separated text file compatible with 
NCBI BLAST m8 and NCBI BLAST+ outfmt 6 formats. One 
cluster/alignment can be found for each line.
}
\examples{
readBlast(system.file(
    "extdata", 
    "packMatches.blast6out", 
    package = "packFinder"
))

}
\references{
For further information, see the NCBI BLAST+ application
documentation and help pages 
(https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
VSEARCH may be downloaded from 
\url{https://github.com/torognes/vsearch}; see 
\url{https://www.ncbi.nlm.nih.gov/pubmed/27781170} 
for further information.
}
\seealso{
code{\link{blastAnalysis}}, code{\link{blastAnnotate}},
code{\link{packAlign}}, 
code{\link{readUc}}, code{\link{packClust}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getTsds.R
\name{getTsds}
\alias{getTsds}
\title{Get Flanking Terminal Site Duplication Sequences}
\usage{
getTsds(tirMatches, Genome, tsdLength, strand = "+", output = "character")
}
\arguments{
\item{tirMatches}{A dataframe containing genomic ranges and names referring 
to TIR sequences or predicted Pack-TYPE transposable 
elements. Should be in the format used by 
\code{\link{packSearch}}.}

\item{Genome}{A \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
object containing sequences referred to in \code{tirMatches}.}

\item{tsdLength}{The length of the TSD region to be retrieved (integer).}

\item{strand}{The strand of the TIR; "+" for forward, "-" for reverse. 
If the TSD sequences of transposable elements are being 
predicted, then this parameter can be left as default 
("+"); if the TSD sequences of TIRs are being found then 
the strand direction must be supplied.}

\item{output}{The type of object to be returned:
\itemize{
    \item output = "DNAStringSet", returns a 
    \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
    object.
    \item output = "character", returns a 
    \code{character} vector (default).
}}
}
\value{
Flanking TSD sequences as a vector of characters, or if 
output is specified as "DNAStringSet", TSD sequences 
will be returned as a 
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
object.
}
\description{
Gets the flanking TSD sequences of TIRs or predicted 
Pack-TYPE transposable elements. A dataframe of these 
elements can be in \code{tirMatches}.
}
\details{
Called by \code{\link{packSearch}}. It is recommended 
to use the general pipeline function 
\code{\link{packSearch}} for identification of potential 
pack elements, which returns TSD sequences as a feature 
of results, however each stage may be called individually.
}
\examples{
data(arabidopsisThalianaRefseq)
data(packMatches)

tsdSeqs <- getTsds(packMatches, arabidopsisThalianaRefseq, 3)

}
\seealso{
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
\code{\link{packSearch}}, \code{tirMatches}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/identifyTirMatches.R
\name{identifyTirMatches}
\alias{identifyTirMatches}
\title{Identify Terminal Inverted Repeat Matches}
\usage{
identifyTirMatches(
  tirSeq,
  Genome,
  mismatch = 0,
  strand = "*",
  tsdLength,
  fixed = TRUE
)
}
\arguments{
\item{tirSeq}{A \code{\link[Biostrings:DNAString-class]{DNAString}} 
object to be searched for.}

\item{Genome}{A \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
object containing the 
\code{\link[Biostrings:DNAString-class]{DNAString}} 
objects to be searched.}

\item{mismatch}{The allowable mismatch between \code{tirSeq} and a 
given slice of \code{Genome}. Includes indels.}

\item{strand}{The directionality of the search string ("+" or "-"). Note
that this does affect the search for tirSeqs, if you wish 
to search the reverse strand you should use the reverse 
complement of your sequence.}

\item{tsdLength}{Integer referring to the length of the flanking TSD region.}

\item{fixed}{Logical that will be passed to the `fixed` argument of 
\code{\link[Biostrings]{matchPattern}}. Determines the behaviour of IUPAC
ambiguity codes when searching for TIR sequences.}
}
\value{
A dataframe, \code{tirMatches}, containing identified 
matches. The dataframe is in the format generated by 
\code{\link{packSearch}}.
}
\description{
Searches a \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
for potential TIRs based on sequence similarity.
}
\details{
Called by \code{\link{packSearch}}. Used by 
\code{\link{packSearch}} as an initial filtering stage. 
\code{\link[Biostrings]{matchPattern}} from Biostrings 
is used for pattern matching. It is recommended to use 
the general pipeline function \code{\link{packSearch}} 
for identification of potential pack elements, however 
each stage may be called individually.
}
\examples{
data(arabidopsisThalianaRefseq)

forwardMatches <- identifyTirMatches(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    tsdLength = 3,
    strand = "+"
)

}
\seealso{
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
\code{\link{packSearch}}, \code{\link[Biostrings]{matchPattern}},
\code{\link[Biostrings:DNAString-class]{DNAString}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packAlign.R
\name{packAlign}
\alias{packAlign}
\title{Global Alignment with VSEARCH}
\usage{
packAlign(
  packMatches,
  Genome,
  identity = 0,
  threads = 1,
  identityDefinition = 2,
  maxWildcards = 0.05,
  saveFolder,
  vSearchPath = "vsearch"
)
}
\arguments{
\item{packMatches}{A dataframe of potential Pack-TYPE transposable elements, 
in the format given by \code{\link{packSearch}}. This 
dataframe is in the format produced by coercing a 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. 
Will be saved as a FASTA file for VSEARCH.}

\item{Genome}{A DNAStringSet object containing sequences referred to in 
\code{packMatches} (the object originally used to predict 
the transposons \code{\link{packSearch}}).}

\item{identity}{The sequence identity of two transposable elements in 
\code{packMatches} required to be grouped into a cluster.}

\item{threads}{The number of threads to be used by VSEARCH.}

\item{identityDefinition}{The pairwise identity definition used by VSEARCH. 
Defaults to 2, the standard VSEARCH definition.}

\item{maxWildcards}{The maximal allowable proportion of wildcards in the 
sequence of each match (defaults to \code{0.05}).}

\item{saveFolder}{The folder to save saveFolder files (uc, blast6out, FASTA)}

\item{vSearchPath}{When the package is run on windows systems, the 
location of the VSEARCH executable file must be 
given; this should be left as default on 
Linux/MacOS systems.}
}
\value{
Saves alignment information, including a \code{uc}, 
\code{blast6out} and a pairwise alignment \code{fasta} 
file, to the specified location. Returns the uc summary 
file generated by the alignment.
}
\description{
A global pairwise alignment of pack-TYPE elements by 
sequence similarity. mIt may be useful to run 
\code{\link{packClust}} to identify groups of similar 
transposable elements, before generating alignments of 
each group.
}
\note{
In order to align sequences using VSEARCH, the 
executable file must first be installed.
}
\examples{
data(arabidopsisThalianaRefseq)
data(packMatches)

# packAlign run on a Linux/MacOS system
\dontrun{
    packAlign(packMatches, Genome)
}

# packAlign run on a Windows system
\dontrun{
    packAlign(packMatches, Genome, 
            vSearchPath = "path/to/vsearch/vsearch.exe")
}

}
\references{
VSEARCH may be downloaded from 
\url{https://github.com/torognes/vsearch}, along with a 
manual documenting the program's parameters. See 
\url{https://www.ncbi.nlm.nih.gov/pubmed/27781170} for 
further information.
}
\seealso{
\code{\link{tirClust}}, \code{\link{packClust}}, 
\code{\link{readBlast}}, \code{\link{readUc}},
\code{\link{filterWildcards}}, \code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getPacksFromGRanges.R
\name{getPacksFromGRanges}
\alias{getPacksFromGRanges}
\title{Retrieve packFinder Results from GRanges Object}
\usage{
getPacksFromGRanges(packGRanges, Genome = NULL, tsdLength = NULL)
}
\arguments{
\item{packGRanges}{\code{link[GenomicRanges:GRanges-class]{GRanges}}
object to be coerced.}

\item{Genome}{(optional) Sequences referred to by \code{packGRanges}.}

\item{tsdLength}{(optional) Length of TSD sequences.}
}
\value{
Dataframe in the format used by \code{\link{packSearch}}. 
If \code{Genome} and \code{tsdLength} are supplied, then 
TSD sequences are retrieved and returned 
as part of the dataframe.
}
\description{
A \code{link[GenomicRanges:GRanges-class]{GRanges}}  
object, potentially 
generated using \code{\link{packSearch}} and 
\code{\link{packsToGRanges}}, can be converted to a 
dataframe. If a GRanges object is supplied without TSD 
information, this can be calculated and appended to the 
final dataframe.
}
\examples{
data(packMatches)

GRangesObject <- packsToGRanges(packMatches)
packMatches <- getPacksFromGRanges(GRangesObject)

}
\seealso{
\code{\link{packsToGRanges}}, 
\code{link[GenomicRanges:GRanges-class]{GRanges}},
\code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packsToFasta.R
\name{packsToFasta}
\alias{packsToFasta}
\title{Save packFinder Results in FASTA Format (.fasta)}
\usage{
packsToFasta(packMatches, file, Genome)
}
\arguments{
\item{packMatches}{taframe containing genomic ranges and names referring 
to sequences to be extracted. Can be obtained from 
\code{\link{packSearch}} or generated from a 
\code{\link[GenomicRanges:GRanges-class]{GRanges}} object, 
after conversion to a dataframe. Must contain the 
following features:
\itemize{
    \item start - the predicted element's start base 
    sequence position.
    \item end - the predicted element's end 
    base sequence position.
    \item seqnames - character string 
    referring to the sequence name in \code{Genome} to 
    which \code{start} and \code{end} refer to.
}}

\item{file}{FASTA file save path.}

\item{Genome}{A DNAStringSet object containing sequences referred to 
in \code{packMatches} (the object originally used to 
predict the transposons \code{\link{packSearch}}).}
}
\value{
Save location of Fasta file.
}
\description{
Saves a dataframe of potential Pack-TYPE elements, 
usually generated via \code{\link{packSearch}}. 
May be retrieved using \code{\link{getPacksFromFasta}}.
}
\examples{
data(arabidopsisThalianaRefseq)
data(packMatches)

packsToFasta(
    packMatches, 
    system.file("extdata", "packMatches.fasta", package = "packFinder"), 
    arabidopsisThalianaRefseq
)

}
\seealso{
\code{\link{getPacksFromFasta}}, \code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packSearch.R
\name{packSearch}
\alias{packSearch}
\title{packFinder Algorithm Pipeline}
\usage{
packSearch(
  tirSeq,
  Genome,
  mismatch = 0,
  elementLength,
  tsdLength,
  tsdMismatch = 0,
  fixed = TRUE
)
}
\arguments{
\item{tirSeq}{A \code{\link[Biostrings:DNAString-class]{DNAString}} 
object containing the TIR sequence to be searched for.}

\item{Genome}{A \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
object to be searched.}

\item{mismatch}{The maximum edit distance to be considered for TIR 
matches (indels + substitions). See 
\code{\link[Biostrings]{matchPattern}} for details.}

\item{elementLength}{The maximum element length to be considered, as a vector 
of two integers. E.g. \code{c(300, 3500)}}

\item{tsdLength}{Integer referring to the length of the flanking TSD region.}

\item{tsdMismatch}{An integer referring to the allowable mismatch 
(substitutions or indels) between a transposon's TSD
sequences. \code{\link[Biostrings]{matchPattern}} from Biostrings 
is used for pattern matching.}

\item{fixed}{Logical that will be passed to the `fixed` argument of 
\code{\link[Biostrings]{matchPattern}}. Determines the behaviour of IUPAC
ambiguity codes when searching for TIR sequences.}
}
\value{
A dataframe, containing elements 
identified by thealgorithm. These may be autonomous or 
pack-TYPE elements. Will contain the following features:
\itemize{
    \item start - the predicted element's start base 
    sequence position.
    \item end - the predicted element's end base 
    sequence position.
    \item seqnames - character string referring to the 
    sequence name in \code{Genome} to which \code{start} 
    and \code{end} refer to.
    \item width - the width of the predicted element.
    \item strand - the strand direction of the 
    transposable element. This will be set to "*" as the 
    \code{packSearch} function does not consider 
    transposons to have a direction - only TIR sequences. 
    Passing the \code{packMatches} dataframe to 
    \code{\link{packClust}} will assign a direction to 
    each predicted Pack-TYPE element.
}

This dataframe is in the format produced by 
coercing a \code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. Downstream 
functions, such as \code{\link{packClust}}, use this 
dataframe to manipulate predicted transposable elements.
}
\description{
General use pipeline function for the Pack-TYPE transposon
finding algorithm.
}
\details{
Finds potential pack-TYPE elements based on:
\itemize{
    \item Similarity of TIR sequence to \code{tirSeq}
    \item Proximity of potential TIR sequences
    \item Directionality of TIR sequences
    \item Similarity of TSD sequences
}

The algorithm finds potential forward and reverse TIR 
sequences using \code{\link{identifyTirMatches}} and 
their associated TSD sequence via \code{\link{getTsds}}. 
The main filtering stage, 
\code{\link{identifyPotentialPackElements}}, filters 
matches to obtain a dataframe of potential PACK elements. 
Note that this pipeline does not consider the 
possibility of discovered elements being autonomous 
elements, so it is recommended to cluster and/or BLAST 
elements for further analysis. Furthermore, only exact TSD 
matches are considered, so supplying long sequences for 
TSD elements may lead to false-negative results.
}
\note{
This algorithm does not consider:
\itemize{
    \item Autonomous elements - autonomous elements will 
    be predicted by this algorithm as there is no BLAST 
    step. It is recommended that, after clustering 
    elements using \code{\link{packClust}}, the user 
    analyses each group to determine which predicted 
    elements are autonomous and which are likely 
    Pack-TYPE elements. Alternatively, databases such as 
    Repbase (\url{https://www.girinst.org/repbase/}) 
    supply annotations for autonomous transposable 
    elements that can be used to filter autonomous matches.
    \item TSD Mismatches - if two TIRs do not have exact 
    matches for their terminal site duplications they 
    will be ignored. Supplying longer TSD sequences will 
    likely lead to a lower false-positive rate, however 
    may also cause a greater rate of false-negative results.
}

Pattern matching is done via \code{\link[Biostrings]{matchPattern}}.
}
\examples{
data(arabidopsisThalianaRefseq)

packMatches <- packSearch(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    elementLength = c(300, 3500),
    tsdLength = 3
)

}
\seealso{
\code{\link{identifyTirMatches}}, \code{\link{getTsds}},
\code{\link{identifyPotentialPackElements}}, \code{\link{packClust}},
\code{\link{packMatches}}, 
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
\code{\link[Biostrings:DNAString-class]{DNAString}},
\code{\link[Biostrings]{matchPattern}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getPackSeqs.R
\name{getPackSeqs}
\alias{getPackSeqs}
\title{Extract Sequences of Pack-TYPE Elements}
\usage{
getPackSeqs(packMatches, Genome, output = "DNAStringSet")
}
\arguments{
\item{packMatches}{A dataframe containing genomic ranges and names referring 
to sequences to be extracted. This dataframe is in the format 
produced by coercing a 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. 

Must contain the following features:
\itemize{
    \item start - the predicted element's start base 
    sequence position.
    \item end - the predicted element's end base 
    sequence position.
    \item seqnames - character string referring to the 
    sequence name in \code{Genome} to which \code{start} 
    and \code{end} refer to.
}}

\item{Genome}{A DNAStringSet object containing sequences referred to 
in \code{packMatches} (the object originally used to 
predict the transposons \code{\link{packSearch}}).}

\item{output}{The type of object to be returned:
\itemize{
    \item output = "DNAStringSet", returns a 
    \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
    object (default).
    \item output = "character", returns a 
    \code{character} vector.
}}
}
\value{
transposon sequences extracted from \code{packMatches}. 
At default returns the sequences as a 
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
or, if \code{output} is set to "character", returns a 
character vector.
}
\description{
Method to quickly extract the sequences of predicted 
Pack-TYPE elements (as created by \code{\link{packSearch}}).
}
\examples{
data(arabidopsisThalianaRefseq)

packMatches <- packSearch(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    elementLength = c(300, 3500),
    tsdLength = 3
)

packSeqs <- getPackSeqs(packMatches, arabidopsisThalianaRefseq)

}
\seealso{
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
\code{\link{packSearch}},
\code{\link[Biostrings:XString-class]{DNAString}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/arabidopsisThalianaRefseq.R
\docType{data}
\name{arabidopsisThalianaRefseq}
\alias{arabidopsisThalianaRefseq}
\title{Arabidopsis thaliana Refseq Genome Chromosome 3 Subset}
\format{
A \code{\link[Biostrings:XStringSet-class]{DNAStringSet}} 
object containing a 
\code{\link[Biostrings:XString-class]{DNAString}} for 
Arabidopsis thaliana's chromosome 3 sequence.
}
\source{
The Arabidopsis thaliana genome was downloaded from the 
NCBI refseq database on 20/SEP/2019, using 
\code{\link[biomartr]{getGenome}}, and chromosome 3 was 
extracted. The genome may also be accessed from the NCBI 
ftp server: \url{ftp://ftp.ncbi.nlm.nih.gov/genomes}.
}
\usage{
data(arabidopsisThalianaRefseq)
}
\description{
The chromosome 3 reference sequence for Arabidopsis 
thaliana as a 
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}}. 
Can be used as a test data set, as in the associated 
introduction vignette. The DNA sequence between bases 
10,500,000 and 14,300,000 was extracted for use in this dataset.
}
\examples{
data(arabidopsisThalianaRefseq)

packMatches <- packSearch(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    elementLength = c(300, 3500),
    tsdLength = 3
)

}
\seealso{
\code{\link[biomartr]{getGenome}},
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
\code{\link[Biostrings:XString-class]{DNAString}},
\code{\link{packSearch}}
}
\author{
Jack Gisby
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blastAnalysis.R
\name{blastAnalysis}
\alias{blastAnalysis}
\title{BLAST Analysis of PackTYPE Elements}
\usage{
blastAnalysis(
  packMatches,
  Genome,
  blastPath,
  protDb = NULL,
  autoDb = NULL,
  minE = 0.001,
  blastTask = "blastn-short",
  maxHits = 100,
  threads = 1,
  saveFolder = NULL,
  tirCutoff = 0
)
}
\arguments{
\item{packMatches}{A dataframe of potential Pack-TYPE transposable elements, 
in the format given by \code{\link{packSearch}}. This 
dataframe is in the format produced by coercing a 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. 
Will be saved as a FASTA file for VSEARCH.}

\item{Genome}{A DNAStringSet object containing sequences referred to 
in \code{packMatches} (the object originally used to 
predict the transposons \code{\link{packSearch}}).}

\item{blastPath}{Path to the BLAST+ executable, or name of 
the BLAST+ application for Linux/MacOS users.}

\item{protDb}{For assigning Pack-TYPE elements. 
Path to the blast database containing nucleotide or protein
sequences to be matched against internal transposon 
sequences. Can be generated 
using BLAST+, or with 
\code{link{makeBlastDb}}.}

\item{autoDb}{For assigning autonomous elements. 
Path to the blast database containing nucleotide or protein
sequences to be matched against internal transposon 
sequences. Can be generated 
using BLAST+, or with 
\code{link{makeBlastDb}}.}

\item{minE}{Blast results with e values greater than
the specified cutoff will be ignored.}

\item{blastTask}{Type of BLAST+ task, defaults to "blastn-short".}

\item{maxHits}{Maximum hits returned by BLAST+ per query.}

\item{threads}{Allowable number of threads to be utilised by BLAST+.}

\item{saveFolder}{Directory to save BLAST+ results in; defaults 
to the working directory.}

\item{tirCutoff}{How many bases to ignore at the terminal ends of the 
transposons to prevent hits to TIR sequences.}
}
\value{
No return value; executes BLAST+ to generate hits 
which are stored in a .blast file in the chosen 
directory.
}
\description{
Run BLAST against user-specified databases of 
non-transposon and transposon-related proteins.
Can be used to classify transposons based on 
their internal sequences.
}
\examples{
\dontrun{
packMatches <- data(packMatches)
Genome <- data(arabidopsisThalianaRefseq)

blastAnalysis(packMatches, Genome, 
    protDb = "C:/data/TAIR10_CDS", 
    autoDb = "C:/data/TAIR10_transposons", 
    blastPath = "C:/blast/bin/blastn.exe")
}

}
\references{
For further information, see the NCBI BLAST+ application
documentation and help pages 
(https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
}
\seealso{
\code{\link{blastAnnotate}},
\code{\link{readBlast}}, \code{\link{packBlast}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packFinder.R
\name{packFinder}
\alias{packFinder}
\title{packFinder: a package for the de novo Annotation 
of Pack-TYPE Transposable Elements}
\description{
Algorithm and tools for in silico pack-TYPE transposon 
discovery. Filters a given genome for properties unique 
to DNA transposons and provides tools for the 
investigation of returned matches.
}
\section{Main Algorithm}{

The goal of packFinder was to implement a simple 
tool for the prediction of potential Pack-TYPE 
elements. packFinder uses the following prior 
knowledge, provided by the user, to detect transposons:

\itemize{
    \item Terminal Inverted Repeat (TIR) Base Sequence
    \item Length of Terminal Site Duplication (TSD)
    \item Length of the Transposon
}

These features provide enough information to detect 
autonomous and pack-TYPE elements. For a transposon 
to be predicted by packFinder its TSD sequences 
must be identical to each other, its forward TIR 
sequence must match the base sequence provided and 
its reverse TIR sequence must match its reverse complement.

Transposons are therefore predicted by searching a 
given genome for these characteristics, and further 
analysis steps can reveal the nature of these elements 
- while the packFinder tool is sensitive for the 
detection of transposons, it does not discriminate 
between autonomous and Pack-TYPE elements. Autonomous 
elements will contain a transposase gene within the 
terminal inverted repeats and tend to be larger than 
their Pack-TYPE counterparts; pack-TYPE elements 
instead capture sections of host genomes. Following 
cluster analysis, BLAST can be used to discern 
which predicted elements are autonomous 
(transposase-containing) and with are true 
Pack-TYPE elements.
}

\section{Workflow}{

An example of a standard workflow can be found using 
\code{browseVignettes(package = "packFinder")}. The 
primary functions include:

\itemize{
    \item packSearch - the packSearch algorithm 
    uses simple pattern matching to detect 
    DNA transposons.
    \item packClust - VSEARCH is used for 
    clustering elements based on sequence 
    similarity. 
}

Having obtained the sequences of transposable
elements in a given genome, it is recommende 
to carry out a BLAST search for each transposon 
cluster. This can identify which elements are 
likely autonomous, and which may be  Pack-TYPE.

The \code{packFinder} functions report the 
position of elements in a given genome using a
dataframe in the format of \code{\link{packMatches}}. 
This dataframe is in the format produced by 
coercing a \code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}.
}

\seealso{
\code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/makeBlastDb.R
\name{makeBlastDb}
\alias{makeBlastDb}
\title{Make Blast Database}
\usage{
makeBlastDb(fastaFile, dbPath, blastPath, dbType = "nucl")
}
\arguments{
\item{fastaFile}{FASTA file containing sequences to generate a BLAST 
database from.}

\item{dbPath}{Path to save the BLAST database to.}

\item{blastPath}{Path/name of BLAST program to use. Name of
the application for Linux/MacOS, absolute 
path for the executable for windows users.}

\item{dbType}{Type of BLAST database to create, e.g. "nucl"
for a nucleotide database.}
}
\value{
No return value; generates a blast database in 
the chosen directory.
}
\description{
Generates a BLAST database to be queried. Required 
for identifying sequences using the BLAST+ 
software.
}
\examples{
\dontrun{
makeBlastDb("genes.fasta", "blastdb.db", "C:/blast.exe")
}

}
\references{
For further information, see the NCBI BLAST+ application
documentation and help pages 
(https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
}
\seealso{
\code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packsToGRanges.R
\name{packsToGRanges}
\alias{packsToGRanges}
\title{Export packFinder Results to a GRanges Object}
\usage{
packsToGRanges(packMatches)
}
\arguments{
\item{packMatches}{A dataframe containing genomic ranges and names 
referring to sequences to be extracted. Can be obtained 
from \code{\link{packSearch}} or generated from a 
\code{\link[GenomicRanges:GRanges-class]{GRanges}} object, 
after conversion to a dataframe. Must contain the 
following features:
\itemize{
    \item start - the predicted element's start 
    base sequence position.
    \item end - the predicted element's end base 
    sequence position.
    \item seqnames - character string referring 
    to the sequence name in \code{Genome} to which 
    \code{start} and \code{end} refer to.
}}
}
\value{
A GRanges object containing the ranges contained in 
\code{packMatches} and additional metadata columns. May 
be easily converted between dataframe and GRanges format 
for use in the \code{packFinder} package and 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
package. Note that most functions in the \code{packFinder} 
package require sequence ranges to be provided in 
dataframe format.
}
\description{
A dataframe containing genomic ranges and names referring 
to sequences to be extracted, likely obtained from 
\code{\link{packSearch}}, can be converted to a GRanges 
object. Can be converted back to a dataframe using 
\code{\link{getPacksFromGRanges}}. Additional features, 
such as clusters and TSD sequences, will be included in 
the object as metadata columns.
}
\examples{
data(packMatches)
packGRanges <- packsToGRanges(packMatches)

}
\seealso{
\code{\link{getPacksFromGRanges}},
\code{link[GenomicRanges:GRanges-class]{GRanges}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blastAnnotate.R
\name{blastAnnotate}
\alias{blastAnnotate}
\title{Functional Annotation of PackTYPE Elements}
\usage{
blastAnnotate(protHits, autoHits, packMatches)
}
\arguments{
\item{protHits}{BLAST results for non-transposon related genes or 
proteins (as a \code{data.frame}). Generated using 
\code{\link{blastAnalysis}}.}

\item{autoHits}{BLAST results for transposon related genes or 
proteins (as a \code{data.frame}). Generated using 
\code{\link{blastAnalysis}}.}

\item{packMatches}{A dataframe of potential Pack-TYPE transposable elements, 
in the format given by \code{\link{packSearch}}. This 
dataframe is in the format produced by coercing a 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. 
Will be saved as a FASTA file for VSEARCH.}
}
\value{
Returns the original \code{packMatches} dataframe, 
with the addition of a "classification" column 
containing one of the following values:
\itemize{
    \item auto - elements that match known 
    transposases or transposon-related proteins 
    are classified as autonomous elements
    \item pack - elements that match other 
    proteins or genic sequences may be classified 
    as Pack-TYPE elements
    \item other - elements that generate no 
    significant hits
}
}
\description{
Uses hits, previously generated using blast, 
to annotate transposon hits. Transposons with 
non-redundant transposase hits are classed as 
autonomous ("auto"), while others are classed as
"other" or "pack" based on whether the element 
has non-redundant hits to other proteins.
}
\note{
Requires that the query ids in the protein and 
autonomous hits match the row names in packMatches.
}
\examples{
data("packMatches")

# read in some protein hits
p <- data.frame(
    query_id = c(2, 3),
    subject_id = c("prot", "hyp")
)

# read in some autonomous hits
a <- data.frame(
    query_id = c(3, 4),
    subject_id = c("transposase", "mutator")
)

blastAnnotate(p, a, packMatches)

}
\references{
For further information, see the NCBI BLAST+ application
documentation and help pages 
(https://www.ncbi.nlm.nih.gov/pubmed/20003500?dopt=Citation).
}
\seealso{
\code{\link{blastAnalysis}},
\code{\link{readBlast}}, \code{\link{packBlast}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tirClust.R
\name{tirClust}
\alias{tirClust}
\title{Analyse TIR Sequences of Pre-clustered Transposable Elements}
\usage{
tirClust(
  packMatches,
  Genome,
  tirLength = 25,
  plot = TRUE,
  plotSavePath = NULL,
  k = 5,
  output = "consensus"
)
}
\arguments{
\item{packMatches}{A dataframe containing genomic ranges and names referring 
to sequences to be extracted. This dataframe is in the format 
produced by coercing a 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. 

Must contain the following features:
\itemize{
    \item start - the predicted element's start base 
    sequence position.
    \item end - the predicted element's end base 
    sequence position.
    \item seqnames - character string referring to the 
    sequence name in \code{Genome} to which \code{start} 
    and \code{end} refer to.
}}

\item{Genome}{A DNAStringSet object containing sequences referred to in 
\code{packMatches} (the object originally used to predict 
the transposons \code{\link{packSearch}}).}

\item{tirLength}{The TIR size to be considered. Consensus sequences will 
be generated based on the first and last \code{tirLength} 
bases of a transposon.}

\item{plot}{Argument specifying whether the TIR consensus sequences 
should be plottted as a dendrogram.}

\item{plotSavePath}{File path for the dendrogram plot. If unspecified, the 
dendrogram plot is not saved.}

\item{k}{The k-mer size to be used for calculating a distance 
matrix between TIR consensus sequences. See 
\code{\link[kmer]{kdistance}}. Larger word sizes will not 
be suitable for longer TIR sequences, due to processing 
time required. Additionally, k must be greater than the 
TIR sequence length.}

\item{output}{Controls the output of \code{tirClust}. If output is 
specified as "consensus", the consensus sequences of each 
TIR cluster will be returned; else, if output is 
specified as "dendrogram", a dendrogram object will be 
returned for creation of customisable plots.}
}
\value{
If \code{output} is specified as "consensus" (default), 
returns a list of consensus sequences for each cluster 
specified in \code{packMatches} as a 
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}}. 
Else if \code{output} is specified as "dendrogram", 
returns a dendrogram object used to create hierarchical 
clustering diagrams.
}
\description{
Takes transposable elements clustered by VSEARCH, 
\code{\link{packClust}}, and produces consensus sequences 
for the terminal inverted repeats of each. Allows for the 
visualisation of TIR similarities between clusters for both 
forward and reverse strands.
}
\examples{
data(arabidopsisThalianaRefseq)
data(packMatches)

tirClust(packMatches, arabidopsisThalianaRefseq)

}
\seealso{
code{\link{packClust}}, code{\link{packAlign}},
\code{\link[kmer]{kdistance}},
\code{\link[Biostrings:XStringSet-class]{DNAStringSet}},
\code{\link[ape]{as.alignment}}, \code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/collapseSeqs.R
\name{collapseSeqs}
\alias{collapseSeqs}
\title{Collapse Overlapping Sequences}
\usage{
collapseSeqs(packMatches, Genome)
}
\arguments{
\item{packMatches}{A dataframe containing genomic ranges and names referring 
to sequences to be extracted. This dataframe is in the format 
produced by coercing a 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}. 

Must contain the following features:
\itemize{
    \item start - the predicted element's start base 
    sequence position.
    \item end - the predicted element's end base 
    sequence position.
    \item seqnames - character string referring to the 
    sequence name in \code{Genome} to which \code{start} 
    and \code{end} refer to.
}}

\item{Genome}{A DNAStringSet object containing sequences referred to 
in \code{packMatches} (the object originally used to 
predict the transposons \code{\link{packSearch}}).}
}
\value{
A set of non-overlapping transposon sequences in the format
of the input dataframe.
}
\description{
The sequences predicted by \code{\link{packSearch}} often 
overlap, which may be due to the presence of closely 
interspersed elements or false TIR identification. 
In such cases, these elements can be combined using 
\code{link[GenomicRanges:GRanges-class]{GRanges}} 
in order to collapse overlapping elements, preventing 
over-estimation of transposon numbers. Also removes 
duplicate elements that have been generated in the 
case of multiple searches.
}
\examples{
data(packMatches)
data(arabidopsisThalianaRefseq)

packMatches$start <- 1
packMatches$end <- 10

collapseSeqs(packMatches, arabidopsisThalianaRefseq)

}
\seealso{
\code{\link{packSearch}},
\code{link[GenomicRanges:GRanges-class]{GRanges}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/packMatches.R
\docType{data}
\name{packMatches}
\alias{packMatches}
\title{Sample packFinder Output}
\format{
A dataframe of 9 obs. and 7 variables.
}
\usage{
data(packMatches)
}
\description{
A sample output from \code{\link{packSearch}} with 
cluster information. This dataframe is in the format produced by 
coercing a \code{link[GenomicRanges:GRanges-class]{GRanges}} 
object to a dataframe: \code{data.frame(GRanges)}.
}
\details{
Was obtained from running \code{\link{packSearch}} 
on the Arabidopsis thaliana chromosome 3 reference 
sequence, followed by clustering using 
\code{\link{packClust}}. Contains the following features:
\itemize{
    \item start - the predicted element's start base 
    sequence position.
    \item end - the predicted element's end base 
    sequence position.
    \item seqnames - character string referring to the 
    sequence name in \code{Genome} to which \code{start} 
    and \code{end} refer to.
}

The dataset was generated as in the example below.
}
\examples{
data(arabidopsisThalianaRefseq)

packMatches <- packSearch(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    elementLength = c(300, 3500),
    tsdLength = 3
)

}
\seealso{
\code{\link{packSearch}}, \code{\link{data.frame}}, 
\code{\link{arabidopsisThalianaRefseq}}
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/getPacksFromCsv.R
\name{getPacksFromCsv}
\alias{getPacksFromCsv}
\title{Retrieve Saved packFinder Results (.csv)}
\usage{
getPacksFromCsv(file)
}
\arguments{
\item{file}{File path to predicted transposons in CSV format.}
}
\value{
Dataframe in the format used by \code{\link{packSearch}}.
}
\description{
Retrieves a dataframe of potential Pack-TYPE elements, 
previously saved using \code{\link{packSearch}} followed 
by \code{\link{packsToCsv}}.
}
\examples{
data(packMatches)

packMatches <- getPacksFromCsv(
    system.file("extdata", "packMatches.csv", package = "packFinder")
)

}
\seealso{
\code{\link{packsToCsv}}, \code{\link[utils]{read.table}}, 
\code{\link{packSearch}}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/identifyPotentialPackElements.R
\name{identifyPotentialPackElements}
\alias{identifyPotentialPackElements}
\title{Pack Element Filtering}
\usage{
identifyPotentialPackElements(
  forwardMatches,
  reverseMatches,
  Genome,
  elementLength,
  tsdMismatch = 0
)
}
\arguments{
\item{forwardMatches}{A dataframe containing genomic ranges and names referring 
to forwards-facing TIR sequences and their respective 
TSD sequences.}

\item{reverseMatches}{A dataframe containing genomic ranges and names referring 
to reverse-facing TIR sequences and their respective 
TSD sequences.}

\item{Genome}{A DNAStringSet object containing the matches referred to 
in \code{forwardMatches} and \code{reverseMatches}}

\item{elementLength}{A vector of two integers containing the minimum and 
maximum transposable element length.}

\item{tsdMismatch}{An integer referring to the allowable mismatch 
(substitutions or indels) between a transposon's TSD
sequences. \code{\link[Biostrings]{matchPattern}} from Biostrings 
is used for pattern matching.}
}
\value{
A dataframe, \code{packMatches}, containing the locations 
of potential Pack-TYPE transposable elements in \code{Genome}.
}
\description{
Primary filtering stage for the \code{packSearch} 
algorithm. Identifies potential Pack-TYPE transposable 
elements based on proximity of matching inverted repeats 
and equality of TSD sequences.
}
\details{
Used by \code{\link{packSearch}} as a primariy filtering 
stage. Identifies matches likely to be transposons based 
on their TIR region, from \code{\link{identifyTirMatches}}, 
and their TSD region, from \code{\link{getTsds}}. It is 
recommended to use the general pipeline function 
\code{\link{packSearch}} for identification of potential 
pack elements, however each stage may be called 
individually. Note that only exact TSD matches are 
considered, so supplying long sequences for TSD elements 
may lead to false-negative results.
}
\examples{
data(arabidopsisThalianaRefseq)

forwardMatches <- identifyTirMatches(
    Biostrings::DNAString("CACTACAA"),
    arabidopsisThalianaRefseq,
    tsdLength = 3,
    strand = "+"
)

reverseMatches <- identifyTirMatches(
    Biostrings::reverseComplement(Biostrings::DNAString("CACTACAA")),
    arabidopsisThalianaRefseq,
    tsdLength = 3,
    strand = "-"
)

packMatches <- identifyPotentialPackElements(
    forwardMatches, 
    reverseMatches,
    arabidopsisThalianaRefseq, 
    c(300, 3500)
)

}
\seealso{
\code{packSearch}
}
\author{
Jack Gisby
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filterWildcards.R
\name{filterWildcards}
\alias{filterWildcards}
\title{Remove Low Quality Sequences}
\usage{
filterWildcards(packMatches, Genome, maxWildcards = 0.05)
}
\arguments{
\item{packMatches}{A dataframe containing genomic ranges 
and names referring to sequences to be extracted.}

\item{Genome}{The original set of sequences used to generate the 
transposons detected by \code{\link{packSearch}}.}

\item{maxWildcards}{The maximal allowable proportion of wildcards in the 
sequence of each match (defaults to \code{0.05}).}
}
\value{
The original dataframe, \code{packMatches}, with 
sequences removed that are found to contain a proportion 
of wildcards ("N") greater than that specified in 
\code{maxWildcards}.
}
\description{
Takes transposable elements detected by 
\code{\link{packSearch}} and removes those with large 
numbers of wildcard ("N") bases. Used by 
\code{\link{packClust}} and \code{\link{packAlign}} to 
remove poor quality sequences that may interfere with 
the quality of sequence alignments.
}
\examples{
data(arabidopsisThalianaRefseq)
data(packMatches)
    
filteredMatches <- filterWildcards(
    packMatches, 
    arabidopsisThalianaRefseq, 
    maxWildcards = 0.05
)

}
\seealso{
\code{\link{packClust}}, \code{\link{packAlign}}, \code{packMatches},
data(arabidopsisThalianaRefseq)
}
\author{
Jack Gisby
}
